#  Copyright (c) 2022 bastien.saltel
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC,                                        \
                abstractmethod
import asyncio
from transitions.core import Machine,                       \
                             EventData
from transitions.extensions.asyncio import AsyncMachine

from galaxy.utils.base import Component,                    \
                              TimestampedState,             \
                              TimestampedAsyncState,        \
                              Configurable
from galaxy.service.service import Manager,                 \
                                   AsyncManager,            \
                                   Service,                 \
                                   AsyncService,            \
                                   LogService,              \
                                   LogAsyncService
from galaxy.service import constant
from galaxy.utils.type import CompId
from galaxy.error.net import NetworkError


class Client(Component, Configurable, ABC):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        Component.__init__(self)
        Configurable.__init__(self)
        self._machine: ClientStateMachine = ClientStateMachine(self)
        self.log: LogService | None = None
        self.is_connected: bool = False

    @abstractmethod
    def _connect(self) -> None:
        raise NotImplementedError("Should implement connect()")

    @abstractmethod
    def _close(self) -> None:
        raise NotImplementedError("Should implement close()")

    def accept(self, visitor):
        visitor.visit(self)

    def __repr__(self) -> str:
        return "<Client(id='{}')>".format(self.id)


class ClientState(TimestampedState):
    """
    classdocs
    """

    def __init__(self, name: str, client: Client) -> None:
        """
        Constructor
        """
        super(ClientState, self).__init__(name=name)
        self.client: Client = client


class ClientNewState(ClientState):
    """
    classdocs
    """

    def __init__(self, client: Client) -> None:
        """
        Constructor
        """
        super(ClientNewState, self).__init__(constant.STATE_NEW, client)


class ClientInitiatedState(ClientState):
    """
    classdocs
    """

    def __init__(self, client: Client) -> None:
        """
        Constructor
        """
        super(ClientInitiatedState, self).__init__(constant.STATE_INIT, client)

    def enter(self, event_data: EventData) -> None:
        self.client.log.logger.debug("The client {} is loading".format(self.client))
        self.client._load()
        self.client.log.logger.debug("The client {} is loaded".format(self.client))
        super(ClientInitiatedState, self).enter(event_data)


class ClientConnectedState(ClientState):
    """
    classdocs
    """

    def __init__(self, client: Client) -> None:
        """
        Constructor
        """
        super(ClientConnectedState, self).__init__(constant.STATE_CONNECTED, client)

    def enter(self, event_data: EventData) -> None:
        self.client.is_connected = False
        self.client.log.logger.debug("The client {} is connecting".format(self.client))
        self.client._connect()
        self.client.log.logger.debug("The client {} is connected".format(self.client))
        self.client.is_connected = True
        super(ClientConnectedState, self).enter(event_data)


class ClientClosedState(ClientState):
    """
    classdocs
    """

    def __init__(self, client: Client) -> None:
        """
        Constructor
        """
        super(ClientClosedState, self).__init__(constant.STATE_CLOSED, client)

    def enter(self, event_data: EventData) -> None:
        self.client.log.logger.debug("The client {} is disconnecting".format(self.client))
        self.client._close()
        self.client.log.logger.debug("The client {} is disconnected".format(self.client))
        super(ClientClosedState, self).enter(event_data)


class ClientTimeoutState(ClientState):
    """
    classdocs
    """

    def __init__(self, client: Client) -> None:
        """
        Constructor
        """
        super(ClientTimeoutState, self).__init__(constant.STATE_TIMEOUT, client)


class ClientStateMachine(object):
    """
    classdocs
    """

    def __init__(self, client: Client) -> None:
        """
        Constructor
        """
        self._client: Client = client
        self._init_states()
        self._init_machine()

    def _init_states(self):
        self.states: dict[str, ClientState] = {
                                               constant.STATE_NEW: ClientNewState(self._client),
                                               constant.STATE_INIT: ClientInitiatedState(self._client),
                                               constant.STATE_CONNECTED: ClientConnectedState(self._client),
                                               constant.STATE_CLOSED: ClientClosedState(self._client),
                                               constant.STATE_TIMEOUT: ClientTimeoutState(self._client)
                                              }
        self._transitions: list[dict[str, str]] = [{
                                                    "trigger": "load",
                                                    "source": constant.STATE_NEW,
                                                    "dest": constant.STATE_INIT
                                                   },
                                                   {
                                                    "trigger": "connect",
                                                    "source": constant.STATE_INIT,
                                                    "dest": constant.STATE_CONNECTED
                                                   },
                                                   {
                                                    "trigger": "close",
                                                    "source": constant.STATE_CONNECTED,
                                                    "dest": constant.STATE_CLOSED
                                                   },
                                                   {
                                                    "trigger": "connect",
                                                    "source": constant.STATE_CLOSED,
                                                    "dest": constant.STATE_CONNECTED
                                                   }]

    def _init_machine(self) -> None:
        self.machine: Machine = Machine(model=self._client,
                                        states=[state for state in self.states.values()],
                                        transitions=self._transitions,
                                        initial=self.states[constant.STATE_NEW])


class AsyncClient(Component, Configurable, ABC):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        Component.__init__(self)
        Configurable.__init__(self)
        self._machine: ClientAsyncStateMachine = ClientAsyncStateMachine(self)
        self.log: LogAsyncService | None = None

        # Boolean in case of connect() method fails in the transition of the State Machine ClientAsyncStateMachine
        # from STATE_INIT or STATE_CLOSED to STATE_CONNECTED (State class ClientConnectedAsyncState).
        # Could be solved if the State Machine can roll the State STATE_CONNECTED back to the initial State STATE_INIT or STATE_CLOSED.
        # This boolean is used in _send_with_retries() of AioHTTPAsyncClient
        self.is_connected: bool = False

    async def _load(self):
        super(AsyncClient, self)._load()

    @abstractmethod
    async def _connect(self) -> None:
        raise NotImplementedError("Should implement connect()")

    @abstractmethod
    async def _close(self) -> None:
        raise NotImplementedError("Should implement close()")

    def accept(self, visitor):
        visitor.visit(self)

    def __repr__(self) -> str:
        return "<AsyncClient(id='{}')>".format(self.id)


class ClientAsyncState(TimestampedAsyncState):
    """
    classdocs
    """

    def __init__(self, name: str, client: AsyncClient) -> None:
        """
        Constructor
        """
        super(ClientAsyncState, self).__init__(name=name)
        self.client: AsyncClient = client


class ClientNewAsyncState(ClientAsyncState):
    """
    classdocs
    """

    def __init__(self, client: AsyncClient) -> None:
        """
        Constructor
        """
        super(ClientNewAsyncState, self).__init__(constant.STATE_NEW, client)


class ClientInitiatedAsyncState(ClientAsyncState):
    """
    classdocs
    """

    def __init__(self, client: AsyncClient) -> None:
        """
        Constructor
        """
        super(ClientInitiatedAsyncState, self).__init__(constant.STATE_INIT, client)

    async def enter(self, event_data: EventData) -> None:
        self.client.log.logger.debug("The client {} is loading".format(self.client))
        await self.client._load()
        self.client.log.logger.debug("The client {} is loaded".format(self.client))
        await super(ClientInitiatedAsyncState, self).enter(event_data)


class ClientConnectedAsyncState(ClientAsyncState):
    """
    classdocs
    """

    def __init__(self, client: AsyncClient) -> None:
        """
        Constructor
        """
        super(ClientConnectedAsyncState, self).__init__(constant.STATE_CONNECTED, client)

    async def enter(self, event_data: EventData) -> None:
        self.client.is_connected = False
        self.client.log.logger.debug("The client {} is connecting".format(self.client))
        try:
            await self.client._connect()
            self.client.log.logger.debug("The client {} is connected".format(self.client))
            self.client.is_connected = True
            await super(ClientConnectedAsyncState, self).enter(event_data)
        except NetworkError as e:
            self.client.log.logger.error("An error occurred : {}".format(str(e)))
            await self.client.close()


class ClientClosedAsyncState(ClientAsyncState):
    """
    classdocs
    """

    def __init__(self, client: AsyncClient) -> None:
        """
        Constructor
        """
        super(ClientClosedAsyncState, self).__init__(constant.STATE_CLOSED, client)

    async def enter(self, event_data: EventData) -> None:
        self.client.log.logger.debug("The client {} is disconnecting".format(self.client))
        await self.client._close()
        self.client.log.logger.debug("The client {} is disconnected".format(self.client))
        await super(ClientClosedAsyncState, self).enter(event_data)


class ClientTimeoutAsyncState(ClientAsyncState):
    """
    classdocs
    """

    def __init__(self, client: AsyncClient) -> None:
        """
        Constructor
        """
        super(ClientTimeoutAsyncState, self).__init__(constant.STATE_TIMEOUT, client)


class ClientAsyncStateMachine(object):
    """
    classdocs
    """

    def __init__(self, client: AsyncClient) -> None:
        """
        Constructor
        """
        self._client: AsyncClient = client
        self._init_states()
        self._init_machine()

    def _init_states(self):
        self.states: dict[str, ClientAsyncState] = {
                                                    constant.STATE_NEW: ClientNewAsyncState(self._client),
                                                    constant.STATE_INIT: ClientInitiatedAsyncState(self._client),
                                                    constant.STATE_CONNECTED: ClientConnectedAsyncState(self._client),
                                                    constant.STATE_CLOSED: ClientClosedAsyncState(self._client),
                                                    constant.STATE_TIMEOUT: ClientTimeoutAsyncState(self._client)
                                                   }
        self._transitions: list[dict[str, str]] = [{
                                                    "trigger": "load",
                                                    "source": constant.STATE_NEW,
                                                    "dest": constant.STATE_INIT
                                                   },
                                                   {
                                                    "trigger": "connect",
                                                    "source": constant.STATE_INIT,
                                                    "dest": constant.STATE_CONNECTED
                                                   },
                                                   {
                                                    "trigger": "close",
                                                    "source": constant.STATE_CONNECTED,
                                                    "dest": constant.STATE_CLOSED
                                                   },
                                                   {
                                                    "trigger": "connect",
                                                    "source": constant.STATE_CLOSED,
                                                    "dest": constant.STATE_CONNECTED
                                                   }]

    def _init_machine(self):
        self.machine: AsyncMachine = AsyncMachine(model=self._client,
                                                  states=[state for state in self.states.values()],
                                                  transitions=self._transitions,
                                                  initial=self.states[constant.STATE_NEW])


class Server(Component, Configurable, ABC):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        Component.__init__(self)
        Configurable.__init__(self)
        self._machine: ServerStateMachine = ServerStateMachine(self)
        self.log: LogService | None = None

    @abstractmethod
    def _start(self) -> None:
        raise NotImplementedError("Should implement _start()")

    @abstractmethod
    def _stop(self) -> None:
        raise NotImplementedError("Should implement _stop()")

    def restart(self) -> None:
        self.stop()
        self.start()

    def accept(self, visitor):
        visitor.visit(self)

    def __repr__(self) -> str:
        return "<Server(id='{}')>".format(self.id)


class ServerState(TimestampedState):
    """
    classdocs
    """

    def __init__(self, name: str, server: Server) -> None:
        """
        Constructor
        """
        super(ServerState, self).__init__(name=name)
        self.server: Server = server


class ServerNewState(ServerState):
    """
    classdocs
    """

    def __init__(self, server: Server) -> None:
        """
        Constructor
        """
        super(ServerNewState, self).__init__(constant.STATE_NEW, server)


class ServerInitiatedState(ServerState):
    """
    classdocs
    """

    def __init__(self, server: Server) -> None:
        """
        Constructor
        """
        super(ServerInitiatedState, self).__init__(constant.STATE_INIT, server)

    def enter(self, event_data: EventData) -> None:
        self.server.log.logger.debug("The server {} is loading".format(self.server))
        self.server._load()
        self.server.log.logger.debug("The server {} is loaded".format(self.server))
        super(ServerInitiatedState, self).enter(event_data)


class ServerRunningState(ServerState):
    """
    classdocs
    """

    def __init__(self, server: Server) -> None:
        """
        Constructor
        """
        super(ServerRunningState, self).__init__(constant.STATE_RUNNING, server)

    def enter(self, event_data: EventData) -> None:
        self.server.log.logger.debug("The server {} is starting".format(self.server))
        self.server._start()
        self.server.log.logger.debug("The server {} is running".format(self.server))
        super(ServerRunningState, self).enter(event_data)


class ServerStoppedState(ServerState):
    """
    classdocs
    """

    def __init__(self, server: Server) -> None:
        """
        Constructor
        """
        super(ServerStoppedState, self).__init__(constant.STATE_STOPPED, server)

    def enter(self, event_data: EventData) -> None:
        self.server.log.logger.debug("The server {} is stopping".format(self.server))
        self.server._stop()
        self.server.log.logger.debug("The server {} is stopped".format(self.server))
        super(ServerStoppedState, self).enter(event_data)


class ServerShutdownState(ServerState):
    """
    classdocs
    """

    def __init__(self, server: Server) -> None:
        """
        Constructor
        """
        super(ServerShutdownState, self).__init__(constant.STATE_SHUTDOWN, server)


class ServerTimeoutState(ServerState):
    """
    classdocs
    """

    def __init__(self, server: Server) -> None:
        """
        Constructor
        """
        super(ServerTimeoutState, self).__init__(constant.STATE_TIMEOUT, server)


class ServerStateMachine(object):
    """
    classdocs
    """

    def __init__(self, server: Server) -> None:
        """
        Constructor
        """
        self._server: Server = server
        self._init_states()
        self._init_machine()

    def _init_states(self) -> None:
        self.states: dict[str, ServerState] = {
                                               constant.STATE_NEW: ServerNewState(self._server),
                                               constant.STATE_INIT: ServerInitiatedState(self._server),
                                               constant.STATE_RUNNING: ServerRunningState(self._server),
                                               constant.STATE_STOPPED: ServerStoppedState(self._server),
                                               constant.STATE_SHUTDOWN: ServerShutdownState(self._server),
                                               constant.STATE_TIMEOUT: ServerTimeoutState(self._server)
                                              }
        self._transitions: list[dict[str, str]] = [{
                                                    "trigger": "load",
                                                    "source": constant.STATE_NEW,
                                                    "dest": constant.STATE_INIT
                                                   },
                                                   {
                                                    "trigger": "start",
                                                    "source": constant.STATE_INIT,
                                                    "dest": constant.STATE_RUNNING
                                                   },
                                                   {
                                                    "trigger": "stop",
                                                    "source": constant.STATE_RUNNING,
                                                    "dest": constant.STATE_STOPPED
                                                   },
                                                   {
                                                    "trigger": "start",
                                                    "source": constant.STATE_STOPPED,
                                                    "dest": constant.STATE_RUNNING
                                                   },
                                                   {
                                                    "trigger": "shutdown",
                                                    "source": constant.STATE_STOPPED,
                                                    "dest": constant.STATE_SHUTDOWN
                                                   }]

    def _init_machine(self) -> None:
        self.machine: Machine = Machine(model=self._server,
                                        states=[state for state in self.states.values()],
                                        transitions=self._transitions,
                                        initial=self.states[constant.STATE_NEW])


class AsyncServer(Component, Configurable, ABC):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        Component.__init__(self)
        Configurable.__init__(self)
        self._machine: ServerAsyncStateMachine = ServerAsyncStateMachine(self)
        self.log: LogAsyncService | None = None

    async def _load(self):
        super(AsyncServer, self)._load()

    @abstractmethod
    async def _start(self) -> None:
        raise NotImplementedError("Should implement _start()")

    @abstractmethod
    async def _stop(self) -> None:
        raise NotImplementedError("Should implement _stop()")

    async def restart(self) -> None:
        await self.stop()
        await self.start()

    def accept(self, visitor):
        visitor.visit(self)

    def __repr__(self) -> str:
        return "<AsyncServer(id='{}')>".format(self.id)


class ServerAsyncState(TimestampedAsyncState):
    """
    classdocs
    """

    def __init__(self, name: str, server: AsyncServer) -> None:
        """
        Constructor
        """
        super(ServerAsyncState, self).__init__(name=name)
        self.server: AsyncServer = server


class ServerNewAsyncState(ServerAsyncState):
    """
    classdocs
    """

    def __init__(self, server: AsyncServer) -> None:
        """
        Constructor
        """
        super(ServerNewAsyncState, self).__init__(constant.STATE_NEW, server)


class ServerInitiatedAsyncState(ServerAsyncState):
    """
    classdocs
    """

    def __init__(self, server: AsyncServer) -> None:
        """
        Constructor
        """
        super(ServerInitiatedAsyncState, self).__init__(constant.STATE_INIT, server)

    async def enter(self, event_data: EventData) -> None:
        # Specific case for Notification Server : Log Service is dependent on this server
        if self.server.log is not None:
            self.server.log.logger.debug("The server {} is loading".format(self.server))
        await self.server._load()
        if self.server.log is not None:
            self.server.log.logger.debug("The server {} is loaded".format(self.server))
        await super(ServerInitiatedAsyncState, self).enter(event_data)


class ServerRunningAsyncState(ServerAsyncState):
    """
    classdocs
    """

    def __init__(self, server: AsyncServer) -> None:
        """
        Constructor
        """
        super(ServerRunningAsyncState, self).__init__(constant.STATE_RUNNING, server)

    async def enter(self, event_data: EventData) -> None:
        if self.server.log is not None:
            self.server.log.logger.debug("The server {} is starting".format(self.server))
        await self.server._start()
        if self.server.log is not None:
            self.server.log.logger.debug("The server {} is running".format(self.server))
        await super(ServerRunningAsyncState, self).enter(event_data)


class ServerStoppedAsyncState(ServerAsyncState):
    """
    classdocs
    """

    def __init__(self, server: AsyncServer) -> None:
        """
        Constructor
        """
        super(ServerStoppedAsyncState, self).__init__(constant.STATE_STOPPED, server)

    async def enter(self, event_data: EventData) -> None:
        if self.server.log is not None:
            self.server.log.logger.debug("The server {} is stopping".format(self.server))
        await self.server._stop()
        if self.server.log is not None:
            self.server.log.logger.debug("The server {} is stopped".format(self.server))
        await super(ServerStoppedAsyncState, self).enter(event_data)


class ServerShutdownAsyncState(ServerAsyncState):
    """
    classdocs
    """

    def __init__(self, server: AsyncServer) -> None:
        """
        Constructor
        """
        super(ServerShutdownAsyncState, self).__init__(constant.STATE_SHUTDOWN, server)


class ServerTimeoutAsyncState(ServerAsyncState):
    """
    classdocs
    """

    def __init__(self, server: AsyncServer) -> None:
        """
        Constructor
        """
        super(ServerTimeoutAsyncState, self).__init__(constant.STATE_TIMEOUT, server)


class ServerAsyncStateMachine(object):
    """
    classdocs
    """

    def __init__(self, server: AsyncServer) -> None:
        """
        Constructor
        """
        self._server: AsyncServer = server
        self._init_states()
        self._init_machine()

    def _init_states(self) -> None:
        self.states: dict[str, ServerAsyncState] = {
                                                    constant.STATE_NEW: ServerNewAsyncState(self._server),
                                                    constant.STATE_INIT: ServerInitiatedAsyncState(self._server),
                                                    constant.STATE_RUNNING: ServerRunningAsyncState(self._server),
                                                    constant.STATE_STOPPED: ServerStoppedAsyncState(self._server),
                                                    constant.STATE_SHUTDOWN: ServerShutdownAsyncState(self._server),
                                                    constant.STATE_TIMEOUT: ServerTimeoutAsyncState(self._server)
                                                   }
        self._transitions: list[dict[str, str]] = [{
                                                    "trigger": "load",
                                                    "source": constant.STATE_NEW,
                                                    "dest": constant.STATE_INIT
                                                   },
                                                   {
                                                    "trigger": "start",
                                                    "source": constant.STATE_INIT,
                                                    "dest": constant.STATE_RUNNING
                                                   },
                                                   {
                                                    "trigger": "stop",
                                                    "source": constant.STATE_RUNNING,
                                                    "dest": constant.STATE_STOPPED
                                                   },
                                                   {
                                                    "trigger": "start",
                                                    "source": constant.STATE_STOPPED,
                                                    "dest": constant.STATE_RUNNING
                                                   },
                                                   {
                                                    "trigger": "shutdown",
                                                    "source": constant.STATE_STOPPED,
                                                    "dest": constant.STATE_SHUTDOWN
                                                   }]

    def _init_machine(self):
        self.machine: AsyncMachine = AsyncMachine(model=self._server,
                                                  states=[state for state in self.states.values()],
                                                  transitions=self._transitions,
                                                  initial=self.states[constant.STATE_NEW])


class NetworkService(Service):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(NetworkService, self).__init__()
        self.clients: dict[str, Client] = {}
        self.servers: dict[str, Server] = {}

    def _load(self) -> None:
        super(NetworkService, self)._load()
        [client.load() for client in self.clients.values()]
        [server.load() for server in self.servers.values()]

    def _start(self) -> None:
        [client.connect() for client in self.clients.values()]
        [server.start() for server in self.servers.values()]

    def _stop(self) -> None:
        [client.close() for client in self.clients.values()]
        [server.stop() for server in self.servers.values()]

    def accept(self, visitor):
        visitor.visit(self)

    def __repr__(self) -> str:
        return "<NetworkService(id='{}')>".format(self.id)


class NetworkAsyncService(AsyncService):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(NetworkAsyncService, self).__init__()
        self.clients: dict[str, AsyncClient] = {}
        self.servers: dict[str, AsyncServer] = {}
        self.loop = None

    async def _load(self) -> None:
        await super(NetworkAsyncService, self)._load()
        client_loads = [client.load() for client in self.clients.values()]
        server_loads = [server.load() for server in self.servers.values()]
        results = await asyncio.gather(*[*client_loads, *server_loads])

        # Hack for initialization of logging service for notif zmq publisher
        for server in [server for server in self.servers.values() if server.log is None]:
            server.log = self.log

    async def _start(self) -> None:
        client_connects = [client.connect() for client in self.clients.values()]
        server_starts = [server.start() for server in self.servers.values()]
        results = await asyncio.gather(*[*client_connects, *server_starts])

    async def _stop(self) -> None:
        client_closes = [client.close() for client in self.clients.values()]
        server_stops = [server.stop() for server in self.servers.values()]
        results = await asyncio.gather(*[*client_closes, *server_stops])

    def accept(self, visitor):
        visitor.visit(self)

    def __repr__(self) -> str:
        return "<NetworkAsyncService(id='{}')>".format(self.id)


class NetworkManager(Manager):
    """
    classdocs
    """
    def __init__(self) -> None:
        """
        Constructor
        """
        super(NetworkManager, self).__init__()

    def __repr__(self) -> str:
        return "<NetworkManager(id='{}')>".format(self.id)


class NetworkAsyncManager(AsyncManager):
    """
    classdocs
    """
    def __init__(self) -> None:
        """
        Constructor
        """
        super(NetworkAsyncManager, self).__init__()

    def __repr__(self) -> str:
        return "<NetworkAsyncManager(id='{}')>".format(self.id)



#  Copyright (c) 2022 bastien.saltel
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from galaxy.utils.pattern import Builder


class Endpoint(object):
    """
    classdocs
    """

    def __init__(self, transport: str, host: str, port: int | None = None) -> None:
        """
        Constructor
        """
        self.transport: str = transport
        self.host: str = host
        self.port: int | None = port
        self.address: str = self._get_address()

    def _get_address(self) -> str:
        if self.port is None:
            return self.host
        return "{}:{}".format(self.host, self.port)

    def _get_endpoint(self) -> str:
        return "{}://{}".format(self.transport, self.address)

    def __str__(self) -> str:
        return self._get_endpoint()

    def __repr__(self) -> str:
        return "<Endpoint(endpoint='{}')>".format(self._get_endpoint())

    def __eq__(self, other: "Endpoint") -> bool:
        return self._get_endpoint() == other._get_endpoint()

    def __ne__(self, other: "Endpoint") -> bool:
        return self._get_endpoint() != other._get_endpoint()


class EndpointBuilder(Builder):
    """
    classdocs
    """

    def __init__(self):
        """
        Constructor
        """
        super(EndpointBuilder, self).__init__()
        self._transport: str | None = None
        self._host: str | None = None
        self._port: int | None = None

    def transport(self, transport: str) -> "EndpointBuilder":
        if transport in ["tcp", "ipc", "inproc", "pgm", "epgm", "vmci"]:
            self._transport = transport
        return self

    def host(self, host: str) -> "EndpointBuilder":
        self._host = host
        return self

    def port(self, port: int) -> "EndpointBuilder":
        self._port = port
        return self

    def from_conf(self, conf: dict) -> "EndpointBuilder":
        if "transport" in conf and conf["transport"] in ["tcp", "ipc", "inproc", "pgm", "epgm", "vmci"]:
            self._transport = conf["transport"]
        if "host" in conf:
            self._host = conf["host"]
        if "port" in conf:
            self._port = conf["port"]
        return self

    def build(self) -> Endpoint:
        return Endpoint(self._transport, self._host, self._port)



#  Copyright (c) 2022 bastien.saltel
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Any
from abc import ABC,                                                    \
                abstractmethod
from urllib.request import Request
from aiohttp.formdata import FormData

from galaxy.net import constant
from galaxy.net.request import HeaderFactory,                           \
                               HeaderBuilder,                           \
                               BodyFactory
from galaxy.net.uri import Uri,                                         \
                           UriPathFactory,                              \
                           UriQueryFactory,                             \
                           UriBuilder
from galaxy.utils.pattern import Builder
from galaxy.utils.base import Component
from galaxy.service.service import LogService,                          \
                                   LogAsyncService


class HTTPHeaderFactory(HeaderFactory, ABC):
    """
    classdocs
    """

    def __init__(self):
        """
        Constructor
        """
        super(HeaderFactory, self).__init__()

    @abstractmethod
    def create(self, **kwargs) -> dict[str, str]:
        raise NotImplementedError("Should implement create()")


class HTTPHeaderBuilder(HeaderBuilder):
    """
    classdocs
    """

    def __init__(self, header_fact: HTTPHeaderFactory | None = None):
        """
        Constructor
        """
        super(HTTPHeaderBuilder, self).__init__()
        self._headers: dict[str, str] = {}
        self.factory: HTTPHeaderFactory | None = header_fact

    def from_conf(self, conf: dict) -> "HTTPHeaderBuilder":
        for header in constant.HTTP_HEADERS:
            conf_header = header.lower().replace("-", "_")
            if conf_header in conf:
                self._headers[header] = conf[conf_header]
        return self

    def build_headers(self, **kwargs) -> "HTTPHeaderBuilder":
        if self.factory is not None:
            self._headers |= self.factory.create(kwargs)
        return self

    def build(self) -> dict[str, str]:
        return self._headers


class HTTPBodyFactory(BodyFactory):
    """
    classdocs
    """

    def __init__(self):
        """
        Constructor
        """
        super(HTTPBodyFactory, self).__init__()

    @abstractmethod
    def create(self, **kwargs) -> str | dict[str, str]:
        raise NotImplementedError("Should implement create()")


class HTTPUriPathFactory(UriPathFactory, ABC):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(HTTPUriPathFactory, self).__init__()
        self.old_path = None

    @abstractmethod
    def create(self, **kwargs) -> str:
        raise NotImplementedError("Should implement create()")


class HTTPUriQueryFactory(UriQueryFactory, ABC):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(HTTPUriQueryFactory, self).__init__()

    @abstractmethod
    def create(self, **kwargs) -> str:
        raise NotImplementedError("Should implement create()")


class HTTPUriBuilder(UriBuilder):
    """
    classdocs
    """

    def __init__(self,
                 path_fact: HTTPUriPathFactory | None = None,
                 query_fact: HTTPUriQueryFactory | None = None) -> None:
        """
        Constructor
        """
        super(HTTPUriBuilder, self).__init__()
        self.log: LogService | None = None
        self.path_fact: HTTPUriPathFactory | None = path_fact
        self.query_fact: HTTPUriQueryFactory | None = query_fact

    def build_path(self, **kwargs) -> "HTTPUriBuilder":
        if self.path_fact is not None:
            self._path = self.path_fact.create(**kwargs)
        return self

    def build_query(self, **kwargs) -> "HTTPUriBuilder":
        if self.query_fact is not None:
            self._query = self.query_fact.create(**kwargs)
        return self


class HTTPRequestBuilder(Component, Builder):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        Component.__init__(self)
        Builder.__init__(self)
        self.log: LogService | LogAsyncService | None = None
        self._uri: Uri | None = None
        self._method: str | None = None
        self._headers: dict[str, str] | None = None
        self._body: str | dict[str, str] | None = None
        self._cookies: Any | None = None

        # Factories
        self.path_fact: HTTPUriPathFactory | None = None
        self.query_fact: HTTPUriQueryFactory | None = None
        self.header_fact: HTTPHeaderFactory | None = None
        self.body_fact: HTTPBodyFactory | None = None

    def uri(self, uri: Uri) -> "HTTPRequestBuilder":
        self._uri = uri
        return self

    def method(self, method: str) -> "HTTPRequestBuilder":
        self._method = method
        return self

    def headers(self, headers: dict[str, str]) -> "HTTPRequestBuilder":
        self._headers = headers
        return self

    def body(self, body: str | dict[str, str] | None) -> "HTTPRequestBuilder":
        self._body = body
        return self

    def cookies(self, cookies: dict[str, str]) -> "HTTPRequestBuilder":
        self._cookies = cookies
        return self

    def from_conf(self, conf: dict, **kwargs) -> "HTTPRequestBuilder":
        # URI
        uri_builder = HTTPUriBuilder(self.path_fact, self.query_fact)
        if "uri" in conf:
            uri_builder.from_conf(conf["uri"])
            if self.path_fact is not None:
                self.path_fact.old_path = conf["uri"]["path"]
        self._uri = uri_builder.build_path(**kwargs).build_query(**kwargs).build()

        # Headers
        header_builder = HTTPHeaderBuilder(self.header_fact)
        if "header" in conf:
            header_builder.from_conf(conf["header"])
        self._headers = header_builder.build_headers(**kwargs).build()

        if "method" in conf:
            self._method = conf["method"]

        # Body
        if self.body_fact is not None:
            self._body = self.body_fact.create(**kwargs)

        return self

    def build(self) -> Request:
        return Request(url=self._uri,
                       data=self._body,
                       headers=self._headers,
                       method=self._method.upper())



#  Copyright (c) 2022 bastien.saltel
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC,                      \
                abstractmethod
import zlib
import gzip

from galaxy.utils.base import Component


class Compressor(Component, ABC):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(Compressor, self).__init__()

    @abstractmethod
    def compress(self, msg: bytes) -> bytes:
        raise NotImplementedError("Should implement compress()")

    @abstractmethod
    def decompress(self, msg: bytes) -> bytes:
        raise NotImplementedError("Should implement decompress()")


class NoCompressor(Compressor):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(NoCompressor, self).__init__()

    def compress(self, data: bytes) -> bytes:
        return data

    def decompress(self, data: bytes) -> bytes:
        return data


class GzipCompressor(Compressor):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(GzipCompressor, self).__init__()

    def compress(self, data: bytes) -> bytes:
        return gzip.compress(data)

    def decompress(self, data: bytes) -> bytes:
        return gzip.decompress(data)


class ZlibCompressor(Compressor):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(ZlibCompressor, self).__init__()

    def compress(self, data: bytes) -> bytes:
        return zlib.compress(data)

    def decompress(self, data: bytes) -> bytes:
        return zlib.decompress(data)



#  Copyright (c) 2022 bastien.saltel
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC,                                                    \
                abstractmethod
import zeep

from galaxy.net.net import Client,                                      \
                           AsyncClient
from galaxy.net.soap.response import SOAPResponseFactory
from galaxy.net.soap.transport import SOAPTransportFactory
from galaxy.utils.base import Component,                                \
                              Configurable
from galaxy.service.service import LogService,                          \
                                   LogAsyncService
from galaxy.net.uri import UriBuilder
from galaxy.perfo.decorator import timed,                               \
                                   async_timed


class SOAPClient(Client, ABC):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """

        super(SOAPClient, self).__init__()
        self.reqreps: dict[str, SOAPRequestReply] | None = None

    def _load(self) -> None:
        super(SOAPClient, self)._load()
        [reqrep._load() for reqrep in self.reqreps.values()]

    @abstractmethod
    def _connect(self) -> None:
        raise NotImplementedError("Should implement connect()")

    @abstractmethod
    def _close(self) -> None:
        raise NotImplementedError("Should implement close()")

    def __repr__(self) -> str:
        return "<SOAPClient(id='{}')>".format(self.id)


class SOAPAsyncClient(AsyncClient):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(SOAPAsyncClient, self).__init__()
        self.reqreps: dict[str, SOAPRequestReply] | None = None

    async def _load(self) -> None:
        await super(SOAPAsyncClient, self)._load()
        [reqrep._load() for reqrep in self.reqreps.values()]

    @abstractmethod
    async def _connect(self) -> None:
        raise NotImplementedError("Should implement connect()")

    @abstractmethod
    async def _close(self) -> None:
        raise NotImplementedError("Should implement close()")

    def __repr__(self) -> str:
        return "<SOAPAsyncClient(id='{}')>".format(self.id)


class ZeepSOAPClient(SOAPClient):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(SOAPClient, self).__init__()
        self.transport_fact: SOAPTransportFactory | None = None
        self.client: zeep.Client | None = None

    @timed
    def _load(self) -> None:
        super(ZeepSOAPClient, self)._load()
        if self.transport_fact is not None:
            self.transport_fact._load()

    @timed
    def _connect(self) -> None:
        transport = self.transport_fact.create()
        uri = UriBuilder().from_conf(self.conf["uri"]).build()
        self.client = zeep.Client(str(uri), transport=transport)
        if self.client is None:
            raise

    @timed
    def _close(self) -> None:
        self.client.transport.close()

    def __repr__(self) -> str:
        return "<ZeepSOAPClient(id='{}')>".format(self.id)


class ZeepSOAPAsyncClient(SOAPAsyncClient):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(ZeepSOAPAsyncClient, self).__init__()
        self.transport_fact: SOAPTransportFactory | None = None
        self.client: zeep.AsyncClient | None

    @async_timed
    async def _load(self) -> None:
        await super(ZeepSOAPAsyncClient, self)._load()
        if self.transport_fact is not None:
            self.transport_fact._load()

    @async_timed
    async def _connect(self) -> None:
        transport = self.transport_fact.create()
        uri = UriBuilder().from_conf(self.conf["uri"]).build()
        self.client = zeep.AsyncClient(str(uri), transport=transport)
        if self.client is None:
            raise

    @async_timed
    async def _close(self) -> None:
        self.client.transport.close()

    def __repr__(self) -> str:
        return "<ZeepSOAPAsyncClient(id='{}')>".format(self.id)


class SOAPRequestReply(Component, Configurable):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        Component.__init__(self)
        Configurable.__init__(self)
        self.log: LogService | LogAsyncService | None = None
        self.resp_fact: SOAPResponseFactory | None = None

    def __repr__(self) -> str:
        return "<SOAPRequestReply(id='{}')>".format(self.id)



#  Copyright (c) 2022 bastien.saltel
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import pysnmp

from galaxy.net.net import Client,              \
                           AsyncClient


class SNMPClient(Client):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(SNMPClient, self).__init__()

    def _connect(self) -> None:
        pass

    def _close(self) -> None:
        pass

    def __repr__(self) -> str:
        return "<SNMPClient(id='{}')>".format(self.id)


class SNMPAsyncClient(AsyncClient):
    """
    classdocs
    """

    def __init__(self):
        """
        Constructor
        """
        super(SNMPAsyncClient, self).__init__()

    async def _connect(self) -> None:
        pass

    async def _close(self) -> None:
        pass

    def __repr__(self) -> str:
        return "<SNMPAsyncClient(id='{}')>".format(self.id)



#  Copyright (c) 2022 bastien.saltel
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC,                                                \
                abstractmethod
from paramiko import client
from typing import Any,                                             \
                   Tuple
from asyncio import Future,                                         \
                    Queue
from asyncssh.connection import SSHClientConnection,                \
                                create_connection
from asyncssh.session import SSHClientSession
from asyncssh.channel import SSHClientChannel
import asyncssh

from galaxy.utils.base import Component,                            \
                              Configurable
from galaxy.kernel.loop import AsyncioLoop
from galaxy.net.net import Client,                                  \
                           AsyncClient,                             \
                           Server,                                  \
                           AsyncServer
from galaxy.net.auth import CredentialBuilder,                      \
                            Credential
from galaxy.net.ssh.transport import SSHTransportFactory
from galaxy.net.ssh.protocol import SSHProtocolFactory
from galaxy.net.ssh.session import SSHSessionFactory
from galaxy.service.log import LogService,                          \
                               LogAsyncService
from galaxy.perfo.decorator import timed,                           \
                                   async_timed


class SSHConnectionFactory(Component, ABC):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(SSHConnectionFactory, self).__init__()
        self.log: LogService | None = None

        # Factory
        self.protocol_fact: SSHProtocolFactory | None = None

    @abstractmethod
    def create(self, client: Any) -> Any:
        raise NotImplementedError("Should implement create()")


class SSHAsyncConnectionFactory(Component, ABC):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(SSHAsyncConnectionFactory, self).__init__()
        self._cred: Credential | None = None
        self.log: LogAsyncService | None = None
        self.loop: AsyncioLoop | None = None
        self.conf: dict[str, Any] | None = None

        # Factory
        self.protocol_fact: SSHProtocolFactory | None = None

    def from_conf(self, conf: dict[str, Any]) -> "SSHAsyncConnectionFactory":
        self.conf = conf

        # Credential
        self._cred = CredentialBuilder().from_conf(self.conf["cred"]).build()

        return self

    @abstractmethod
    async def create(self, client: Any) -> Any:
        raise NotImplementedError("Should implement create()")


class AsyncSSHAsyncConnectionFactory(SSHAsyncConnectionFactory):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(AsyncSSHAsyncConnectionFactory, self).__init__()

    async def create(self, client: Any) -> Tuple[SSHClientConnection, asyncssh.SSHClient]:
        # options parameter do not work, parameters are overridden in method "connect", file asyncssh/connection.py, line 8743
        # new_options = await SSHServerConnectionOptions.construct(options, config=config, **kwargs)
        # With respect of the comment, options are used when starting the reverse-direction of an SSH server

        # opts = SSHClientConnectionOptions()
        # opts.prepare(self.loop.loop,
        #              host=self.conf["host"],
        #              port=self.conf["port"],
        #              username=self._cred.username,
        #              password=self._cred.password,
        #              passphrase=self._cred.passphrase,
        #              client_keys=[self._cred.private_key] if self._cred.private_key is not None else None,
        #              known_hosts=None)
        closed = Future()
        queue = Queue()
        return await create_connection(lambda: self.protocol_fact.create(client, closed, queue),
                                       host=self.conf["host"],
                                       port=self.conf["port"],
                                       username=self._cred.username,
                                       password=self._cred.password,
                                       passphrase=self._cred.passphrase,
                                       client_keys=[self._cred.private_key] if self._cred.private_key is not None else None,
                                       known_hosts=None)


class SSHAsyncSessionFactory(Component, ABC):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(SSHAsyncSessionFactory, self).__init__()

        # Factory
        self.session_fact: SSHSessionFactory | None = None

    @abstractmethod
    async def create(self, transport: Any, cmd: str) -> Any:
        raise NotImplementedError("Should implement create()")


class AsyncSSHAsyncSessionFactory(SSHAsyncSessionFactory):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(AsyncSSHAsyncSessionFactory, self).__init__()

    async def create(self, transport: asyncssh.SSHClient, cmd: str) -> Tuple[SSHClientChannel, SSHClientSession]:
        closed = Future()
        queue = Queue()
        return await transport.create_session(lambda: self.session_fact.create_session(closed, queue), cmd)


class SSHClient(Client):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(SSHClient, self).__init__()

    @abstractmethod
    def _connect(self) -> None:
        raise NotImplementedError("Should implement _connect()")

    @abstractmethod
    def exec(self, cmd: str) -> None:
        raise NotImplementedError("Should implement exec()")

    @abstractmethod
    def _close(self) -> None:
        raise NotImplementedError("Should implement _close()")

    def __repr__(self) -> str:
        return "<SSHClient(id='{}')>".format(self.id)


class ParamikoSSHClient(SSHClient):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(ParamikoSSHClient, self).__init__()
        self.transport_fact: SSHTransportFactory | None = None
        self.client: client.SSHClient | None = None

    @timed
    def _load(self) -> None:
        super(ParamikoSSHClient, self)._load()

    def _connect(self) -> None:
        if self.client is None:
            self.client = client.SSHClient()
        cred = CredentialBuilder().from_conf(self.conf["cred"]).build()
        self.client.connect(hostname=self.conf["host"],
                            port=self.conf["port"],
                            username=cred.username,
                            password=cred.password)

    def exec(self, cmd: str) -> None:
        pass

    def _close(self) -> None:
        return self.client.close()

    def __repr__(self) -> str:
        return "<ParamikoSSHClient(id='{}')>".format(self.id)


class SSHAsyncClient(AsyncClient, ABC):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(SSHAsyncClient, self).__init__()
        self.loop: AsyncioLoop | None = None

        # Factory
        self.conn_fact: SSHAsyncConnectionFactory | None = None
        self.session_fact: SSHAsyncSessionFactory | None = None

    @async_timed
    async def _load(self) -> None:
        await super(SSHAsyncClient, self)._load()

    @abstractmethod
    async def _connect(self) -> None:
        raise NotImplementedError("Should implement _connect()")

    @abstractmethod
    async def exec(self, cmd: str) -> None:
        raise NotImplementedError("Should implement exec()")

    @abstractmethod
    async def _close(self) -> None:
        pass

    def __repr__(self) -> str:
        return "<SSHAsyncClient(id='{}')>".format(self.id)


class AsyncSSHAsyncClient(SSHAsyncClient):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(AsyncSSHAsyncClient, self).__init__()
        self.protocol: SSHClientConnection | None = None
        self.transport: asyncssh.SSHClient | None = None
        self.session: SSHClientSession | None = None

    @async_timed
    async def _load(self) -> None:
        await super(AsyncSSHAsyncClient, self)._load()

    @async_timed
    async def _connect(self) -> None:
        self.transport, self.protocol = await self.conn_fact.from_conf(self.conf).create()

    @async_timed
    async def exec(self, cmd: str) -> None:
        async with self.transport:
            chan, self.session = await self.session_fact.create(self.transport, cmd)
            await chan.wait_closed()

    @async_timed
    async def _close(self) -> None:
        pass

    def __repr__(self) -> str:
        return "<AsyncSSHAsyncClient(id='{}')>".format(self.id)


class SSHServer(Server, ABC):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(SSHServer, self).__init__()

    @abstractmethod
    def _start(self) -> None:
        raise NotImplementedError("Should implement _start()")

    @abstractmethod
    def _stop(self) -> None:
        raise NotImplementedError("Should implement _stop()")

    def __repr__(self) -> str:
        return "<SSHServer(id='{}')>".format(self.id)


class SSHAsyncServer(AsyncServer, ABC):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(SSHAsyncServer, self).__init__()

    @abstractmethod
    async def _start(self) -> None:
        raise NotImplementedError("Should implement _start()")

    @abstractmethod
    async def _stop(self) -> None:
        raise NotImplementedError("Should implement _stop()")

    def __repr__(self) -> str:
        return "<SSHAsyncServer(id='{}')>".format(self.id)



#  Copyright (c) 2022-2025 Sucden Financial Limited.
#
#  Written by bastien.saltel.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC,                                            \
                abstractmethod
from typing import Any
import asyncio
from asyncio import Future,                                     \
                    Queue
from asyncssh.client import SSHClient
from asyncssh.connection import SSHClientConnection

from galaxy.kernel.loop import AsyncioLoop
from galaxy.utils.base import Component
from galaxy.service.service import LogService,                  \
                                   LogAsyncService
from galaxy.net.ssh.session import AsyncSSHSessionFactory,      \
                                   AsyncSSHSession


class SSHProtocolFactory(Component, ABC):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(SSHProtocolFactory, self).__init__()
        self.log: LogService | LogAsyncService | None = None

    @abstractmethod
    def create(self, client: Any) -> Any:
        raise NotImplementedError("Should implement create()")


class AsyncSSHProtocol(SSHClient):
    """
    classdocs
    """

    def __init__(self, client: Any, on_close: Future, log: LogService | LogAsyncService, queue: Queue | None = None) -> None:
        """
        Constructor
        """
        super(AsyncSSHProtocol, self).__init__()
        self.client: Any = client
        self.transport: SSHClientConnection | None = None
        self.session: AsyncSSHSession | None = None
        self.on_close: Future = on_close
        self.queue: Queue | None = queue
        self.log = log

    def connection_made(self, transport: SSHClientConnection) -> None:
        self.transport = transport
        self.log.logger.debug("The client {} is connected".format(self.client))

    def connection_lost(self, ex: Exception | None):
        if ex is not None:
            self.log.logger.debug("The connection of the client {} has been lost : {}".format(self.client, ex))
            self.on_close.set_result(ex)
        else:
            self.log.logger.debug("The client {} has been disconnected".format(self.client))

    def auth_completed(self) -> None:
        pass


class AsyncSSHProtocolFactory(Component):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(AsyncSSHProtocolFactory, self).__init__()
        self.log: LogService | LogAsyncService | None = None

        # Factory
        self.session_fact: AsyncSSHSessionFactory | None = None

    def create(self, client: Any, on_close: Future, queue: Queue | None = None) -> AsyncSSHProtocol:
        protocol = AsyncSSHProtocol(client, on_close, self.log, queue)
        protocol.session = self.session_fact.create(on_close, queue)
        return protocol



#  Copyright (c) 2022-2023 Sucden Financial Limited.
#
#  Written by bastien.saltel.
#
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from uuid import UUID
import zmq
from zmq import ZMQError,                                   \
                Context,                                    \
                Socket,                                     \
                Poller
from zmq.eventloop.zmqstream import ZMQStream
from asyncio import Future,                                 \
                    Queue
from aiozmq.core import create_zmq_connection
from aiozmq.interface import ZmqTransport,                  \
                             ZmqProtocol
from collections.abc import Iterable
import errno
from logging import getLogger,                              \
                    Formatter,                              \
                    StreamHandler,                          \
                    BASIC_FORMAT
from abc import ABC,                                        \
                abstractmethod
from threading import Thread, Event
import time
from ipykernel.iostream import IOPubThread
from jupyter_client.blocking import BlockingKernelClient
from jupyter_client.session import Session
from IPython.core.profiledir import ProfileDir
from tornado import ioloop
from typing import Any,                                     \
                   Tuple

from galaxy.net.net import Server,                          \
                           AsyncServer,                     \
                           Client,                          \
                           AsyncClient
from galaxy.net import constant
from galaxy.net.endpoint import EndpointBuilder,            \
                                Endpoint
from galaxy.net.zmq.protocol import ZmqProtocolFactory
from galaxy.data.serial import Serializer,                  \
                               NoSerializer
from galaxy.net.compression import Compressor,              \
                                   NoCompressor
from galaxy.kernel.loop import AsyncioLoop
from galaxy.utils.base import Component,                    \
                              Configurable
from galaxy.service.log import LogService,                  \
                               LogAsyncService
from galaxy.perfo.decorator import timed,                   \
                                   async_timed


class ZmqConnectionFactory(Component, ABC):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(ZmqConnectionFactory, self).__init__()
        self.log: LogService | None = None

        # Factory
        self.protocol_fact: ZmqProtocolFactory | None = None

    @abstractmethod
    def create(self, zmq_type: str) -> Any:
        raise NotImplementedError("Should implement create()")


class ZmqAsyncConnectionFactory(Component, ABC):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(ZmqAsyncConnectionFactory, self).__init__()
        self._endpoint: Endpoint | None = None
        self.log: LogAsyncService | None = None
        self.loop: AsyncioLoop | None = None

        # Factory
        self.protocol_fact: ZmqProtocolFactory | None = None

    def from_conf(self, conf: dict[str, Any]) -> "ZmqAsyncConnectionFactory":
        # Endpoint
        self._endpoint = EndpointBuilder().from_conf(conf).build()

        return self

    @abstractmethod
    async def create(self, zmq_type: str) -> Any:
        raise NotImplementedError("Should implement create()")


class AioZmqAsyncServerConnectionFactory(ZmqAsyncConnectionFactory):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(AioZmqAsyncServerConnectionFactory, self).__init__()

    async def create(self, zmq_type: str) -> Tuple[ZmqTransport, ZmqProtocol]:
        closed = Future()
        queue = Queue()
        return await create_zmq_connection(lambda: self.protocol_fact.create(closed, queue),
                                           constant.ZMQ_TYPES[zmq_type],
                                           bind=str(self._endpoint),
                                           loop=self.loop.loop)


class AioZmqAsyncClientConnectionFactory(ZmqAsyncConnectionFactory):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(AioZmqAsyncClientConnectionFactory, self).__init__()

    async def create(self, zmq_type: str) -> Tuple[ZmqTransport, ZmqProtocol]:
        closed = Future()
        queue = Queue()
        return await create_zmq_connection(lambda: self.protocol_fact.create(closed, queue),
                                           constant.ZMQ_TYPES[zmq_type],
                                           connect=str(self._endpoint),
                                           loop=self.loop.loop)


class ZmqServer(Server, ABC):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(ZmqServer, self).__init__()
        self.serializer: Serializer = NoSerializer()
        self.compressor: Compressor = NoCompressor()

        # Factory
        self.conn_fact: ZmqConnectionFactory | None = None

    @timed
    def _load(self) -> None:
        super(ZmqServer, self)._load()
        if self.conn_fact is not None:
            self.conn_fact._load()

    @abstractmethod
    def _start(self) -> None:
        raise NotImplementedError("Should implement _start()")

    @abstractmethod
    def _stop(self) -> None:
        raise NotImplementedError("Should implement _stop()")

    @abstractmethod
    def send(self, data: Iterable[bytes] | Iterable[bytearray] | Iterable[memoryview]) -> None:
        raise NotImplementedError("Should implement send()")

    def __repr__(self) -> str:
        return "<ZmqServer(id='{}')>".format(self.id)


class ZmqAsyncServer(AsyncServer, ABC):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(ZmqAsyncServer, self).__init__()
        self.protocol: ZmqProtocol | None = None
        self.transport: ZmqTransport | None = None
        self.serializer: Serializer = NoSerializer()
        self.compressor: Compressor = NoCompressor()

        # Factory
        self.conn_fact: ZmqAsyncConnectionFactory | None = None

    @async_timed
    async def _load(self) -> None:
        await super(ZmqAsyncServer, self)._load()

    @abstractmethod
    async def _start(self) -> None:
        raise NotImplementedError("Should implement _start()")

    @abstractmethod
    async def _stop(self) -> None:
        raise NotImplementedError("Should implement _stop()")

    @abstractmethod
    def send(self, data: Iterable[bytes] | Iterable[bytearray] | Iterable[memoryview]) -> None:
        raise NotImplementedError("Should implement send()")

    def __repr__(self):
        return "<ZmqAsyncServer(id='{}')>".format(self.id)


class AioZmqAsyncServer(ZmqAsyncServer):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(AioZmqAsyncServer, self).__init__()
        self.protocol: ZmqProtocol | None = None
        self.transport: ZmqTransport | None = None

    @async_timed
    async def _load(self) -> None:
        await super(AioZmqAsyncServer, self)._load()

    @async_timed
    async def _start(self) -> None:
        self.transport, self.protocol = await self.conn_fact.from_conf(self.conf["endpoint"]).create(self.conf["zmq_type"])

        if "socket_options" in self.conf:
            for option, value in self.conf["socket_options"].items():
                if option == "SUBSCRIBE":
                    self.transport.setsockopt(constant.ZMQ_SOCKET_OPTIONS(option), value.bytes)
                elif option in ["IDENTITY", "ROUTING_ID"]:
                    self.transport.setsockopt(constant.ZMQ_SOCKET_OPTIONS[option], UUID(value).bytes)
                else:
                    self.transport.setsockopt(constant.ZMQ_SOCKET_OPTIONS(option), value)

    @async_timed
    async def _stop(self) -> None:
        self.transport.close()

    @timed
    def send(self, data: Iterable[bytes] | Iterable[bytearray] | Iterable[memoryview]) -> None:
        self.transport.write(data)

    def __repr__(self):
        return "<AioZmqAsyncServer(id='{}')>".format(self.id)


class ZmqClient(Client, ABC):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(ZmqClient, self).__init__()
        self.serializer: Serializer = NoSerializer()
        self.compressor: Compressor = NoCompressor()

        # Factory
        self.conn_fact: ZmqConnectionFactory | None = None

    @timed
    def _load(self) -> None:
        super(ZmqClient, self)._load()

    @abstractmethod
    def _connect(self) -> None:
        raise NotImplementedError("Should implement _connect()")

    @abstractmethod
    async def _close(self) -> None:
        raise NotImplementedError("Should implement _close()")

    @abstractmethod
    def send(self, data: Iterable[bytes] | Iterable[bytearray] | Iterable[memoryview]) -> None:
        raise NotImplementedError("Should implement send()")

    def __repr__(self) -> str:
        return "<ZmqClient(id='{}')>".format(self.id)


class ZmqAsyncClient(AsyncClient, ABC):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(ZmqAsyncClient, self).__init__()
        self.serializer: Serializer = NoSerializer()
        self.compressor: Compressor = NoCompressor()

        # Factory
        self.conn_fact: ZmqAsyncConnectionFactory | None = None

    @async_timed
    async def _load(self) -> None:
        await super(ZmqAsyncClient, self)._load()

    @abstractmethod
    def _connect(self) -> None:
        raise NotImplementedError("Should implement _connect()")

    @abstractmethod
    async def _close(self) -> None:
        raise NotImplementedError("Should implement _close()")

    @abstractmethod
    def send(self, data: Iterable[bytes] | Iterable[bytearray] | Iterable[memoryview]) -> None:
        raise NotImplementedError("Should implement send()")

    def __repr__(self) -> str:
        return "<ZmqAsyncClient(id='{}')>".format(self.id)


class AioZmqAsyncClient(ZmqAsyncClient):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(AioZmqAsyncClient, self).__init__()
        self.protocol: ZmqProtocol | None = None
        self.transport: ZmqTransport | None = None

    @async_timed
    async def _load(self) -> None:
        await super(AioZmqAsyncClient, self)._load()

    @async_timed
    async def _connect(self) -> None:
        self.transport, self.protocol = await self.conn_fact.from_conf(self.conf["endpoint"]).create(self.conf["zmq_type"])

        if "socket_options" in self.conf:
            for option, value in self.conf["socket_options"].items():
                if option == "SUBSCRIBE":
                    self.transport.setsockopt(constant.ZMQ_SOCKET_OPTIONS[option], bytes(value, "utf-8"))
                elif option in ["IDENTITY", "ROUTING_ID"]:
                    self.transport.setsockopt(constant.ZMQ_SOCKET_OPTIONS[option], UUID(value).bytes)
                else:
                    self.transport.setsockopt(constant.ZMQ_SOCKET_OPTIONS[option], value)

    @async_timed
    async def _close(self) -> None:
        self.transport.close()

    @timed
    def send(self, data: Iterable[bytes] | Iterable[bytearray] | Iterable[memoryview]) -> None:
        self.transport.write(data)

    def __repr__(self) -> str:
        return "<AioZmqAsyncClient(id='{}')>".format(self.id)

# class ZmqRouter(ZmqServer, ABC):
#     """
#     classdocs
#     """
#
#     def __init__(self) -> None:
#         """
#         Constructor
#         """
#         super(ZmqRouter, self).__init__()
#         self.socket_type: int = zmq.ROUTER
#         self.stream: ZMQStream | None = None
#
#     def _start(self) -> None:
#         super(ZmqRouter, self)._start()
#         self.stream = ZMQStream(self.socket)
#         if hasattr(zmq, 'ROUTER_HANDOVER'):
#             # set router-handover to workaround zeromq reconnect problems
#             # in certain rare circumstances
#             # see ipython/ipykernel#270 and zeromq/libzmq#2892
#             self.socket.router_handover = 1
#
#     def _stop(self) -> None:
#         super(ZmqRouter, self)._stop()
#
#     def __repr__(self) -> str:
#         return "<ZmqRouter(id='{}')>".format(self.id)
#
#
# class ZmqAsyncRouter(ZmqAsyncServer, ABC):
#     """
#     classdocs
#     """
#
#     def __init__(self) -> None:
#         """
#         Constructor
#         """
#         super(ZmqAsyncRouter, self).__init__()
#         self.socket_type: int = zmq.ROUTER
#         self.stream: ZMQStream | None = None
#
#     async def _start(self) -> None:
#         await super(ZmqAsyncRouter, self)._start()
#         self.stream = ZMQStream(self.socket)
#         if hasattr(zmq, 'ROUTER_HANDOVER'):
#             # set router-handover to workaround zeromq reconnect problems
#             # in certain rare circumstances
#             # see ipython/ipykernel#270 and zeromq/libzmq#2892
#             self.socket.router_handover = 1
#
#     async def _stop(self) -> None:
#         await super(ZmqAsyncRouter, self)._stop()
#
#     def __repr__(self) -> str:
#         return "<ZmqAsyncRouter(id='{}')>".format(self.id)
#
#
# class HeartbeatZmqRouterThread(Thread):
#     """
#     classdocs
#     """
#
#     def __init__(self, socket) -> None:
#         """
#         Constructor
#         """
#         super(HeartbeatZmqRouterThread, self).__init__()
#         self.socket: Socket = socket
#
#     def run(self) -> None:
#         while True:
#             try:
#                 zmq.device(zmq.QUEUE, self.socket, self.socket)
#             except zmq.ZMQError as e:
#                 if e.errno == errno.EINTR:
#                     # signal interrupt, resume heartbeat
#                     continue
#                 elif e.errno == zmq.ETERM:
#                     # context terminated, close socket and exit
#                     try:
#                         self.socket.close()
#                     except zmq.ZMQError:
#                         # suppress further errors during cleanup
#                         # this shouldn't happen, though
#                         pass
#                     break
#                 elif e.errno == zmq.ENOTSOCK:
#                     # socket closed elsewhere, exit
#                     break
#                 else:
#                     raise
#             else:
#                 break
#
#
# class HeartbeatZmqRouter(ZmqRouter):
#     """
#     classdocs
#     """
#
#     def __init__(self) -> None:
#         """
#         Constructor
#         """
#         super(HeartbeatZmqRouter, self).__init__()
#         self.thread: HeartbeatZmqRouterThread | None = None
#
#     def _start(self) -> None:
#         super(HeartbeatZmqRouter, self)._start()
#         self.thread = HeartbeatZmqRouterThread(self.socket)
#         self.log.logger.debug("The heartbeat ROUTER socket is opened on {}".format(self.conf["port"]))
#         self.thread.start()
#
#     def _stop(self) -> None:
#         super(HeartbeatZmqRouter, self)._stop()
#         self.thread.join()
#
#     def __repr__(self) -> str:
#         return "<HeartbeatZmqRouter(id='{}')>".format(self.id)
#
#
# class HeartbeatZmqAsyncRouter(ZmqAsyncRouter):
#     """
#     classdocs
#     """
#
#     def __init__(self) -> None:
#         """
#         Constructor
#         """
#         super(HeartbeatZmqAsyncRouter, self).__init__()
#
#     async def _start(self) -> None:
#         await super(HeartbeatZmqAsyncRouter, self)._start()
#         self.log.logger.debug("Heartbeat REP Channel on port: {}".format(self.conf["port"]))
#
#     async def _stop(self) -> None:
#         await super(HeartbeatZmqAsyncRouter, self)._stop()
#
#     def __repr__(self) -> str:
#         return "<HeartbeatZmqAsyncRouter(id='{}')>".format(self.id)
#
#
# class HeartbeatZmqService(NetworkService):
#     """
#     classdocs
#     """
#
#     def __init__(self) -> None:
#         """
#         Constructor
#         """
#         super(HeartbeatZmqService, self).__init__()
#         self.pub_server = None
#         self.router_client = None
#
#     def _load(self) -> None:
#         super(HeartbeatZmqService, self)._load()
#         self.router_client.stream.on_recv(self.handle_pong)
#
#     def _start(self) -> None:
#         self.tic = time.time()
#         io_loop = ioloop.IOLoop.current()
#         self.caller = io_loop.PeriodicCallback(self.beat, self.period)
#         self.caller.start()
#
#     def _stop(self) -> None:
#         super(HeartbeatZmqRouter, self)._stop()
#         self.thread.join()
#
#     def __repr__(self) -> str:
#         return "<HeartbeatZmqRouter(id='{}')>".format(self.id)
#
#
# class ShellZmqRouter(ZmqRouter):
#     """
#     classdocs
#     """
#
#     def __init__(self) -> None:
#         """
#         Constructor
#         """
#         super(ShellZmqRouter, self).__init__()
#
#     def _start(self) -> None:
#         super(ShellZmqRouter, self)._start()
#
#     def _stop(self) -> None:
#         super(ShellZmqRouter, self)._stop()
#
#     def __repr__(self) -> str:
#         return "<ShellZmqRouter(id='{}')>".format(self.id)
#
#
# class ShellZmqAsyncRouter(ZmqAsyncRouter):
#     """
#     classdocs
#     """
#
#     def __init__(self) -> None:
#         """
#         Constructor
#         """
#         super(ShellZmqAsyncRouter, self).__init__()
#
#     async def _start(self) -> None:
#         await super(ShellZmqAsyncRouter, self)._start()
#
#     async def _stop(self) -> None:
#         await super(ShellZmqAsyncRouter, self)._stop()
#
#     def __repr__(self) -> str:
#         return "<ShellZmqRouter(id='{}')>".format(self.id)
#
#
# class StdinZmqRouter(ZmqRouter):
#     """
#     classdocs
#     """
#
#     def __init__(self) -> None:
#         """
#         Constructor
#         """
#         super(StdinZmqRouter, self).__init__()
#
#     def _start(self) -> None:
#         super(StdinZmqRouter, self)._start()
#
#     def _stop(self) -> None:
#         pass
#
#     def __repr__(self) -> str:
#         return "<StdinZmqRouter(id='{}')>".format(self.id)
#
#
# class StdinZmqAsyncRouter(ZmqAsyncRouter):
#     """
#     classdocs
#     """
#
#     def __init__(self) -> None:
#         """
#         Constructor
#         """
#         super(StdinZmqAsyncRouter, self).__init__()
#
#     async def _start(self) -> None:
#         super(StdinZmqAsyncRouter, self)._start()
#
#     async def _stop(self) -> None:
#         pass
#
#     def __repr__(self) -> str:
#         return "<StdinZmqAsyncRouter(id='{}')>".format(self.id)
#
#
# class ControlZmqRouter(ZmqRouter):
#     """
#     classdocs
#     """
#
#     def __init__(self) -> None:
#         """
#         Constructor
#         """
#         super(ControlZmqRouter, self).__init__()
#
#     def _start(self) -> None:
#         super(ControlZmqRouter, self)._start()
#
#     def _stop(self) -> None:
#         pass
#
#     def __repr__(self) -> str:
#         return "<ControlZmqRouter(id='{}')>".format(self.id)
#
#
# class ControlZmqAsyncRouter(ZmqAsyncRouter):
#     """
#     classdocs
#     """
#
#     def __init__(self) -> None:
#         """
#         Constructor
#         """
#         super(ControlZmqAsyncRouter, self).__init__()
#
#     async def _start(self) -> None:
#         super(ControlZmqAsyncRouter, self)._start()
#
#     async def _stop(self) -> None:
#         pass
#
#     def __repr__(self) -> str:
#         return "<ControlZmqAsyncRouter(id='{}')>".format(self.id)
#
#
# class CommandZmqRouter(ZmqRouter):
#     """
#     classdocs
#     """
#
#     def __init__(self) -> None:
#         """
#         Constructor
#         """
#         super(CommandZmqRouter, self).__init__()
#
#     def _start(self) -> None:
#         super(CommandZmqRouter, self)._start()
#
#     def _stop(self) -> None:
#         pass
#
#     def __repr__(self) -> str:
#         return "<CommandZmqRouter(id='{}')>".format(self.id)
#
#
# class CommandZmqAsyncRouter(ZmqAsyncRouter):
#     """
#     classdocs
#     """
#
#     def __init__(self) -> None:
#         """
#         Constructor
#         """
#         super(CommandZmqAsyncRouter, self).__init__()
#
#     async def _start(self) -> None:
#         await super(CommandZmqAsyncRouter, self)._start()
#
#     async def _stop(self) -> None:
#         await super(CommandZmqAsyncRouter, self)._stop()
#
#     def __repr__(self) -> str:
#         return "<CommandZmqAsyncRouter(id='{}')>".format(self.id)
#
#
# class ZmqPublisher(ZmqServer, ABC):
#     """
#     classdocs
#     """
#
#     def __init__(self) -> None:
#         """
#         Constructor
#         """
#         super(ZmqPublisher, self).__init__()
#         self.socket_type = zmq.PUB
#
#     def _start(self) -> None:
#         super(ZmqPublisher, self)._start()
#
#     def _stop(self) -> None:
#         super(ZmqPublisher, self)._stop()
#
#     def __repr__(self) -> str:
#         return "<ZmqPublisher(id='{}')>".format(self.id)
#
#
# class ZmqAsyncPublisher(ZmqAsyncServer, ABC):
#     """
#     classdocs
#     """
#
#     def __init__(self):
#         """
#         Constructor
#         """
#         super(ZmqAsyncPublisher, self).__init__()
#         self.socket_type = zmq.PUB
#
#     async def _start(self):
#         await super(ZmqAsyncPublisher, self)._start()
#
#     async def _stop(self):
#         await super(ZmqAsyncPublisher, self)._stop()
#
#     def __repr__(self):
#         return "<ZmqAsyncPublisher(id='{}')>".format(self.id)
#
#
# class IOZmqPublisher(ZmqPublisher):
#     """
#     classdocs
#     """
#
#     def __init__(self) -> None:
#         """
#         Constructor
#         """
#         super(IOZmqPublisher, self).__init__()
#         self.thread = None
#
#     def _start(self) -> None:
#         super(IOZmqPublisher, self)._start()
#         self._configure_tornado_logger()
#         self.thread = IOPubThread(self.socket, pipe=True)
#         self.thread.start()
#         self.socket = self.thread.background_socket
#
#     def _configure_tornado_logger(self) -> None:
#         logger = getLogger('tornado')
#         handler = StreamHandler()
#         formatter = Formatter(BASIC_FORMAT)
#         handler.setFormatter(formatter)
#         logger.addHandler(handler)
#
#     def _stop(self) -> None:
#         super(IOZmqPublisher, self)._stop()
#         self.thread.stop()
#         self.thread.close()
#
#     def __repr__(self) -> str:
#         return "<IOZmqPublisher(id='{}')>".format(self.id)
#
#
# class IOZmqAsyncPublisher(ZmqAsyncPublisher):
#     """
#     classdocs
#     """
#
#     def __init__(self) -> None:
#         """
#         Constructor
#         """
#         super(IOZmqAsyncPublisher, self).__init__()
#         self.thread = None
#
#     async def _start(self) -> None:
#         await super(IOZmqAsyncPublisher, self)._start()
#
#     async def _stop(self) -> None:
#         await super(IOZmqAsyncPublisher, self)._stop()
#
#     def __repr__(self) -> str:
#         return "<IOZmqAsyncPublisher(id='{}')>".format(self.id)
#
#
# class NotifZmqPublisher(ZmqPublisher):
#     """
#     classdocs
#     """
#
#     def __init__(self) -> None:
#         """
#         Constructor
#         """
#         super(NotifZmqPublisher, self).__init__()
#
#     def _start(self) -> None:
#         super(NotifZmqPublisher, self)._start()
#
#     def _stop(self) -> None:
#         super(NotifZmqPublisher, self)._stop()
#
#     def __repr__(self) -> str:
#         return "<NotifZmqPublisher(id='{}')>".format(self.id)
#
#
# class NotifZmqAsyncPublisher(ZmqAsyncPublisher):
#     """
#     classdocs
#     """
#
#     def __init__(self) -> None:
#         """
#         Constructor
#         """
#         super(NotifZmqAsyncPublisher, self).__init__()
#
#     async def _start(self) -> None:
#         await super(NotifZmqAsyncPublisher, self)._start()
#
#     async def _stop(self) -> None:
#         await super(NotifZmqAsyncPublisher, self)._stop()
#
#     def __repr__(self) -> str:
#         return "<NotifZmqAsyncPublisher(id='{}')>".format(self.id)
#
#
# class MonitorZmqPublisher(ZmqPublisher):
#     """
#     classdocs
#     """
#
#     def __init__(self) -> None:
#         """
#         Constructor
#         """
#         super(MonitorZmqPublisher, self).__init__()
#
#     def _start(self) -> None:
#         super(MonitorZmqPublisher, self)._start()
#
#     def _stop(self) -> None:
#         super(MonitorZmqPublisher, self)._stop()
#
#     def __repr__(self) -> str:
#         return "<MonitorZmqPublisher(id='{}')>".format(self.id)
#
#
# class MonitorZmqAsyncPublisher(ZmqAsyncPublisher):
#     """
#     classdocs
#     """
#
#     def __init__(self) -> None:
#         """
#         Constructor
#         """
#         super(MonitorZmqAsyncPublisher, self).__init__()
#
#     async def _start(self) -> None:
#         await super(MonitorZmqAsyncPublisher, self)._start()
#
#     async def _stop(self) -> None:
#         await super(MonitorZmqAsyncPublisher, self)._stop()
#
#     def __repr__(self) -> str:
#         return "<MonitorZmqAsyncPublisher(id='{}')>".format(self.id)
#
#
# class DataZmqPublisher(ZmqPublisher):
#     """
#     classdocs
#     """
#
#     def __init__(self) -> None:
#         """
#         Constructor
#         """
#         super(DataZmqPublisher, self).__init__()
#
#     def _start(self) -> None:
#         super(DataZmqPublisher, self)._start()
#
#     def _stop(self) -> None:
#         super(DataZmqPublisher, self)._stop()
#
#     def __repr__(self) -> str:
#         return "<DataZmqPublisher(id='{}')>".format(self.id)
#
#
# class DataZmqAsyncPublisher(ZmqAsyncPublisher):
#     """
#     classdocs
#     """
#
#     def __init__(self) -> None:
#         """
#         Constructor
#         """
#         super(DataZmqAsyncPublisher, self).__init__()
#
#     async def _start(self) -> None:
#         await super(DataZmqAsyncPublisher, self)._start()
#
#     async def _stop(self) -> None:
#         await super(DataZmqAsyncPublisher, self)._stop()
#
#     def __repr__(self) -> str:
#         return "<DataZmqAsyncPublisher(id='{}')>".format(self.id)
#
#
# class ZmqDealer(ZmqClient, ABC):
#     """
#     classdocs
#     """
#
#     def __init__(self) -> None:
#         """
#         Constructor
#         """
#         super(ZmqDealer, self).__init__()
#         self.socket_type = zmq.DEALER
#
#     def _connect(self) -> None:
#         super(ZmqDealer, self)._connect()
#
#     def _disconnect(self) -> None:
#         super(ZmqDealer, self)._disconnect()
#
#     def _close(self) -> None:
#         super(ZmqDealer, self)._close()
#
#     def __repr__(self) -> str:
#         return "<ZmqDealer(id='{}')>".format(self.id)
#
#
# class ZmqAsyncDealer(ZmqAsyncClient, ABC):
#     """
#     classdocs
#     """
#
#     def __init__(self) -> None:
#         """
#         Constructor
#         """
#         super(ZmqAsyncDealer, self).__init__()
#         self.socket_type = zmq.DEALER
#
#     async def _connect(self) -> None:
#         await super(ZmqAsyncDealer, self)._connect()
#
#     async def _disconnect(self) -> None:
#         await super(ZmqAsyncDealer, self)._disconnect()
#
#     async def _close(self) -> None:
#         await super(ZmqAsyncDealer, self)._close()
#
#     def __repr__(self) -> str:
#         return "<ZmqAsyncDealer(id='{}')>".format(self.id)
#
#
# class ShellZmqDealer(ZmqDealer):
#     """
#     classdocs
#     """
#
#     def __init__(self) -> None:
#         """
#         Constructor
#         """
#         super(ShellZmqDealer, self).__init__()
#
#     def _connect(self) -> None:
#         super(ShellZmqDealer, self)._connect()
#
#     def _disconnect(self) -> None:
#         super(ShellZmqDealer, self)._disconnect()
#
#     def _close(self) -> None:
#         super(ShellZmqDealer, self)._close()
#
#     def __repr__(self) -> str:
#         return "<ShellZmqDealer(id='{}')>".format(self.id)
#
#
# class ShellZmqAsyncDealer(ZmqAsyncDealer):
#     """
#     classdocs
#     """
#
#     def __init__(self) -> None:
#         """
#         Constructor
#         """
#         super(ShellZmqAsyncDealer, self).__init__()
#
#     async def _connect(self) -> None:
#         await super(ShellZmqAsyncDealer, self)._connect()
#
#     async def _disconnect(self) -> None:
#         await super(ShellZmqAsyncDealer, self)._disconnect()
#
#     async def _close(self) -> None:
#         await super(ShellZmqAsyncDealer, self)._close()
#
#     def __repr__(self) -> str:
#         return "<ShellZmqAsyncDealer(id='{}')>".format(self.id)
#
#
# class StdinZmqDealer(ZmqDealer):
#     """
#     classdocs
#     """
#
#     def __init__(self) -> None:
#         """
#         Constructor
#         """
#         super(StdinZmqDealer, self).__init__()
#
#     def _connect(self) -> None:
#         super(StdinZmqDealer, self)._connect()
#
#     def _disconnect(self) -> None:
#         super(StdinZmqDealer, self)._disconnect()
#
#     def _close(self) -> None:
#         super(StdinZmqDealer, self)._close()
#
#     def __repr__(self) -> str:
#         return "<StdinZmqDealer(id='{}')>".format(self.id)
#
#
# class StdinZmqAsyncDealer(ZmqAsyncDealer):
#     """
#     classdocs
#     """
#
#     def __init__(self) -> None:
#         """
#         Constructor
#         """
#         super(StdinZmqAsyncDealer, self).__init__()
#
#     async def _connect(self) -> None:
#         await super(StdinZmqAsyncDealer, self)._connect()
#
#     async def _disconnect(self) -> None:
#         await super(StdinZmqAsyncDealer, self)._disconnect()
#
#     async def _close(self) -> None:
#         await super(StdinZmqAsyncDealer, self)._close()
#
#     def __repr__(self) -> str:
#         return "<StdinZmqAsyncDealer(id='{}')>".format(self.id)
#
#
# class ControlZmqDealer(ZmqDealer):
#     """
#     classdocs
#     """
#
#     def __init__(self) -> None:
#         """
#         Constructor
#         """
#         super(ControlZmqDealer, self).__init__()
#
#     def _connect(self) -> None:
#         super(ControlZmqDealer, self)._connect()
#
#     def _disconnect(self) -> None:
#         super(ControlZmqDealer, self)._disconnect()
#
#     def _close(self) -> None:
#         super(ControlZmqDealer, self)._close()
#
#     def __repr__(self) -> str:
#         return "<ControlZmqDealer(id='{}')>".format(self.id)
#
#
# class ControlZmqAsyncDealer(ZmqAsyncDealer):
#     """
#     classdocs
#     """
#
#     def __init__(self) -> None:
#         """
#         Constructor
#         """
#         super(ControlZmqAsyncDealer, self).__init__()
#
#     async def _connect(self) -> None:
#         await super(ControlZmqAsyncDealer, self)._connect()
#
#     async def _disconnect(self) -> None:
#         await super(ControlZmqAsyncDealer, self)._disconnect()
#
#     async def _close(self) -> None:
#         await super(ControlZmqAsyncDealer, self)._close()
#
#     def __repr__(self) -> str:
#         return "<ControlZmqDealer(id='{}')>".format(self.id)
#
#
# class StdoutZmqDealer(ZmqDealer):
#     """
#     classdocs
#     """
#
#     def __init__(self) -> None:
#         """
#         Constructor
#         """
#         super(StdoutZmqDealer, self).__init__()
#
#     def _connect(self) -> None:
#         super(StdoutZmqDealer, self)._connect()
#
#     def _disconnect(self) -> None:
#         super(StdoutZmqDealer, self)._disconnect()
#
#     def _close(self) -> None:
#         super(StdoutZmqDealer, self)._close()
#
#     def __repr__(self) -> str:
#         return "<StdoutZmqDealer(id='{}')>".format(self.id)
#
#
# class StdoutZmqAsyncDealer(ZmqAsyncDealer):
#     """
#     classdocs
#     """
#
#     def __init__(self) -> None:
#         """
#         Constructor
#         """
#         super(StdoutZmqAsyncDealer, self).__init__()
#
#     async def _connect(self) -> None:
#         await super(StdoutZmqAsyncDealer, self)._connect()
#
#     async def _disconnect(self) -> None:
#         await super(StdoutZmqAsyncDealer, self)._disconnect()
#
#     async def _close(self) -> None:
#         await super(StdoutZmqAsyncDealer, self)._close()
#
#     def __repr__(self) -> str:
#         return "<StdoutZmqAsyncDealer(id='{}')>".format(self.id)
#
#
# class ZmqSubscriber(ZmqClient, ABC):
#     """
#     classdocs
#     """
#
#     def __init__(self) -> None:
#         """
#         Constructor
#         """
#         super(ZmqSubscriber, self).__init__()
#         self.socket_type = zmq.SUB
#
#     def _connect(self) -> None:
#         super(ZmqSubscriber, self)._connect()
#         self.socket.setsockopt(zmq.SUBSCRIBE, b'')
#
#     def _disconnect(self) -> None:
#         super(ZmqSubscriber, self)._disconnect()
#
#     def _close(self) -> None:
#         super(ZmqSubscriber, self)._close()
#
#     def __repr__(self) -> str:
#         return "<ZmqSubscriber(id='{}')>".format(self.id)
#
#
# class ZmqAsyncSubscriber(ZmqAsyncClient, ABC):
#     """
#     classdocs
#     """
#
#     def __init__(self) -> None:
#         """
#         Constructor
#         """
#         super(ZmqAsyncSubscriber, self).__init__()
#         self.socket_type = zmq.SUB
#
#     async def _connect(self) -> None:
#         await super(ZmqAsyncSubscriber, self)._connect()
#         self.socket.setsockopt(zmq.SUBSCRIBE, b'')
#
#     async def _disconnect(self) -> None:
#         await super(ZmqAsyncSubscriber, self)._disconnect()
#
#     async def _close(self) -> None:
#         await super(ZmqAsyncSubscriber, self)._close()
#
#     def __repr__(self) -> str:
#         return "<ZmqAsyncSubscriber(id='{}')>".format(self.id)
#
#
# class IOZmqSubscriber(ZmqSubscriber):
#     """
#     classdocs
#     """
#
#     def __init__(self) -> None:
#         """
#         Constructor
#         """
#         super(IOZmqSubscriber, self).__init__()
#
#     def _connect(self) -> None:
#         super(IOZmqSubscriber, self)._connect()
#
#     def _disconnect(self) -> None:
#         super(IOZmqSubscriber, self)._disconnect()
#
#     def _close(self) -> None:
#         super(IOZmqSubscriber, self)._close()
#
#     def __repr__(self) -> str:
#         return "<IOZmqSubscriber(id='{}')>".format(self.id)
#
#
# class IOZmqAsyncSubscriber(ZmqAsyncSubscriber):
#     """
#     classdocs
#     """
#
#     def __init__(self) -> None:
#         """
#         Constructor
#         """
#         super(IOZmqAsyncSubscriber, self).__init__()
#
#     async def _connect(self) -> None:
#         await super(IOZmqAsyncSubscriber, self)._connect()
#
#     async def _disconnect(self) -> None:
#         await super(IOZmqAsyncSubscriber, self)._disconnect()
#
#     async def _close(self) -> None:
#         await super(IOZmqAsyncSubscriber, self)._close()
#
#     def __repr__(self) -> str:
#         return "<IOZmqAsyncSubscriber(id='{}')>".format(self.id)
#
#
# class ZmqRequest(ZmqClient, ABC):
#     """
#     classdocs
#     """
#
#     def __init__(self) -> None:
#         """
#         Constructor
#         """
#         super(ZmqRequest, self).__init__()
#         self.socket_type = zmq.REQ
#
#     def _connect(self) -> None:
#         super(ZmqRequest, self)._connect()
#
#     def _disconnect(self) -> None:
#         super(ZmqRequest, self)._disconnect()
#
#     def _close(self) -> None:
#         super(ZmqRequest, self)._close()
#
#     def __repr__(self) -> str:
#         return "<ZmqRequest(id='{}')>".format(self.id)
#
#
# class ZmqAsyncRequest(ZmqAsyncClient, ABC):
#     """
#     classdocs
#     """
#
#     def __init__(self) -> None:
#         """
#         Constructor
#         """
#         super(ZmqAsyncRequest, self).__init__()
#         self.socket_type = zmq.REQ
#
#     async def _connect(self) -> None:
#         await super(ZmqAsyncRequest, self)._connect()
#
#     async def _disconnect(self) -> None:
#         await super(ZmqAsyncRequest, self)._disconnect()
#
#     async def _close(self) -> None:
#         await super(ZmqAsyncRequest, self)._close()
#
#     def __repr__(self) -> str:
#         return "<ZmqAsyncRequest(id='{}')>".format(self.id)
#
#
# class HeartbeatZmqRequestThread(Thread):
#     """
#     classdocs
#     """
#
#     def __init__(self, req) -> None:
#         """
#         Constructor
#         """
#         super(HeartbeatZmqRequestThread, self).__init__()
#         self.req = req
#         self._exit: Event = Event()
#
#     def _poll(self, start_time) -> None:
#         """poll for heartbeat replies until we reach self.time_to_dead.
#
#         Ignores interrupts, and returns the result of poll(), which
#         will be an empty list if no messages arrived before the timeout,
#         or the event tuple if there is a message to receive.
#         """
#
#         until_dead = self.req.time_to_dead - (time.time() - start_time)
#         # ensure poll at least once
#         until_dead = max(until_dead, 1e-3)
#         events = []
#         while True:
#             try:
#                 events = self.req.poller.poll(1000 * until_dead)
#             except ZMQError as e:
#                 if e.errno == errno.EINTR:
#                     # ignore interrupts during heartbeat
#                     # this may never actually happen
#                     until_dead = self.req.time_to_dead - (time.time() - start_time)
#                     until_dead = max(until_dead, 1e-3)
#                     pass
#                 else:
#                     raise
#             except Exception:
#                 #if self._exiting:
#                 #    break
#                 #else:
#                 raise
#             else:
#                 break
#         return events
#
#     def run(self) -> None:
#         self.req._beating = True
#
#         #while self._running:
#         while True:
#             #if self._pause:
#                 # just sleep, and skip the rest of the loop
#             #    self._exit.wait(self.time_to_dead)
#             #    continue
#
#             since_last_heartbeat = 0.0
#             # no need to catch EFSM here, because the previous event was
#             # either a recv or connect, which cannot be followed by EFSM
#             self.req.socket.send(b'ping')
#             request_time = time.time()
#             ready = self._poll(request_time)
#             if ready:
#                 self.req._beating = True
#                 # the poll above guarantees we have something to recv
#                 msg = self.req.socket.recv()
#                 print(msg)
#                 # sleep the remainder of the cycle
#                 remainder = self.req.time_to_dead - (time.time() - request_time)
#                 if remainder > 0:
#                     self._exit.wait(remainder)
#                 continue
#             else:
#                 # nothing was received within the time limit, signal heart failure
#                 self.req._beating = False
#                 since_last_heartbeat = time.time() - request_time
#                 self.call_handlers(since_last_heartbeat)
#                 # and close/reopen the socket, because the REQ/REP cycle has been broken
#                 self.req.disconnect()
#                 self.req.connect()
#                 continue
#
#     def call_handlers(self, since_last_heartbeat) -> None:
#         """This method is called in the ioloop thread when a message arrives.
#
#         Subclasses should override this method to handle incoming messages.
#         It is important to remember that this method is called in the thread
#         so that some logic must be done to ensure that the application level
#         handlers are called in the application thread.
#         """
#         pass
#
#
# class HeartbeatZmqRequest(ZmqRequest):
#     """
#     classdocs
#     """
#
#     def __init__(self) -> None:
#         """
#         Constructor
#         """
#         super(HeartbeatZmqRequest, self).__init__()
#         self.poller: Poller = Poller()
#         self._beating: bool = False
#         self.thread: HeartbeatZmqRequestThread | None = None
#         self.time_to_dead: int = 1
#
#     def _connect(self) -> None:
#         super(HeartbeatZmqRequest, self)._connect()
#         self.poller.register(self.socket, zmq.POLLIN)
#         self.thread = HeartbeatZmqRequestThread(self)
#         self.log.logger.debug("Heartbeat REQ Channel on port: {}".format(self.conf["port"]))
#         self.thread.start()
#
#     def is_beating(self) -> bool:
#         return self._beating
#
#     def __repr__(self) -> str:
#         return "<HeartbeatZmqRequest(id='{}')>".format(self.id)
#
#
# class HeartbeatZmqAsyncRequest(ZmqAsyncRequest):
#     """
#     classdocs
#     """
#
#     def __init__(self) -> None:
#         """
#         Constructor
#         """
#         super(HeartbeatZmqAsyncRequest, self).__init__()
#         self.poller: Poller = Poller()
#         self._beating: bool = False
#         self.thread: HeartbeatZmqRequestThread | None = None
#         self.time_to_dead: int = 1
#
#     async def _connect(self) -> None:
#         await super(HeartbeatZmqAsyncRequest, self)._connect()
#         self.poller.register(self.socket, zmq.POLLIN)
#         self.thread = HeartbeatZmqRequestThread(self)
#         self.log.logger.debug("Heartbeat REQ Channel on port: {}".format(self.conf["port"]))
#         self.thread.start()
#
#     def is_beating(self) -> bool:
#         return self._beating
#
#     def __repr__(self) -> str:
#         return "<HeartbeatZmqAsyncRequest(id='{}')>".format(self.id)
#
#
# class JupyterClientMeta(type(NetworkService), type(BlockingKernelClient)):
#     pass
#
#
# class JupyterClientService(NetworkService, BlockingKernelClient, metaclass=JupyterClientMeta):
#     """
#     classdocs
#     """
#
#     def __init__(self) -> None:
#         """
#         Constructor
#         """
#         NetworkService.__init__(self)
#         BlockingKernelClient.__init__(self)
#         self.log = None
#
#         self.profile_dir: ProfileDir = ProfileDir()
#         self.user_ns = None
#
#     def _load(self) -> None:
#         NetworkService._load(self)
#         self._create_session()
#
#     def _create_session(self) -> None:
#         self.session = Session(parent=self)
#         self.session.session = self.conf["session"]
#         self.session.bsession = self.conf["session"].encode("utf-8")
#         self.session.key = self.conf["key"].encode("utf-8")
#         self.session.signature_scheme = self.conf["signature_scheme"]
#         self.session.username = self.conf["username"]
#
#         for client in self.clients.values():
#             client.session = self.session
#         self.clients["shell"].identity = self.session.bsession
#         self.clients["stdin"].identity = self.session.bsession
#         self.clients["ctrl"].identity = self.session.bsession
#
#     def _start(self) -> None:
#         NetworkService._start(self)
#         self._inject_client()
#
#     def _inject_client(self) -> None:
#         self._shell_channel = self.clients["shell"]
#         self._iopub_channel = self.clients["iopub"]
#         self._stdin_channel = self.clients["stdin"]
#         self._hb_channel = self.clients["hb"]
#         self._control_channel = self.clients["ctrl"]
#
#
# class JupyterAsyncClientService(NetworkAsyncService, AsyncClient, metaclass=JupyterClientMeta):
#     """
#     classdocs
#     """
#
#     def __init__(self) -> None:
#         """
#         Constructor
#         """
#         NetworkAsyncService.__init__(self)
#         AsyncClient.__init__(self)
#         self.log = None
#
#         self.profile_dir = ProfileDir()
#         self.user_ns = None
#
#     async def _load(self) -> None:
#         NetworkAsyncService._load(self)
#         self._create_session()
#
#     async def _create_session(self) -> None:
#         self.session = Session(parent=self)
#         self.session.session = self.conf["session"]
#         self.session.bsession = self.conf["session"].encode("utf-8")
#         self.session.key = self.conf["key"].encode("utf-8")
#         self.session.signature_scheme = self.conf["signature_scheme"]
#         self.session.username = self.conf["username"]
#
#         for client in self.clients.values():
#             client.session = self.session
#         self.clients["shell"].identity = self.session.bsession
#         self.clients["stdin"].identity = self.session.bsession
#         self.clients["ctrl"].identity = self.session.bsession
#
#     async def _start(self) -> None:
#         NetworkAsyncService._start(self)
#         self._inject_client()
#
#     async def _inject_client(self) -> None:
#         self._shell_channel = self.clients["shell"]
#         self._iopub_channel = self.clients["iopub"]
#         self._stdin_channel = self.clients["stdin"]
#         self._hb_channel = self.clients["hb"]
#         self._control_channel = self.clients["ctrl"]



#  Copyright (c) 2022 bastien.saltel
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC,                                    \
                abstractmethod
from typing import Any
from socket import socket
from paramiko.transport import Transport

from galaxy.utils.base import Component,                \
                              Configurable
from galaxy.service.service import LogService,          \
                                   LogAsyncService
from galaxy.net.socket import SocketFactory


class SSHTransportFactory(Component, ABC):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(SSHTransportFactory, self).__init__()
        self.log: LogService | LogAsyncService | None = None

    @abstractmethod
    def create(self) -> Any:
        raise NotImplementedError("Should implement create()")


class ParamikoSSHTransportFactory(SSHTransportFactory):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(ParamikoSSHTransportFactory, self).__init__()
        self.sock: socket | None = None

    def create(self) -> Transport:
        sock = SocketFactory().create(self.conf["host"], self.conf["port"])
        return Transport(sock, gss_kex=False)



#  Copyright (c) 2022 bastien.saltel
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Any
from abc import ABC,                                                    \
                abstractmethod

from galaxy.net.response import ResponseFactory


class SOAPResponseFactory(ResponseFactory, ABC):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(SOAPResponseFactory, self).__init__()

    @abstractmethod
    def create(self, data: str) -> Any:
        raise NotImplementedError("Should implement create()")


class DefaultSOAPResponseFactory(SOAPResponseFactory):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(DefaultSOAPResponseFactory, self).__init__()

    def create(self, data: str) -> str:
        self.log.log.debug(data)
        return data



#  Copyright (c) 2022 bastien.saltel
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import stat
import threading
import socket
import sys
import traceback
import os
import asyncio
import logging
from abc import ABC,                                                    \
                abstractmethod
import paramiko
from paramiko import sftp_client,                                       \
                     sftp_server
from paramiko.server import ServerInterface
from paramiko.sftp_attr import SFTPAttributes
from pysftp import CnOpts
from typing import Callable,                                            \
                   Awaitable,                                           \
                   Any
from asyncssh.connection import listen
from asyncssh.channel import SSHServerChannel
from asyncssh import sftp
from asyncssh.misc import ConnectionLost
from asyncssh.sftp import SFTPNoConnection
import asyncssh
from tenacity.asyncio import AsyncRetrying
from tenacity.retry import retry_if_result,                             \
                           retry_if_exception_type
from tenacity.stop import stop_after_attempt
from tenacity.wait import wait_fixed
from tenacity import RetryError
from tenacity.before import before_log
from tenacity.after import after_log

from galaxy.service import constant
from galaxy.net.net import Client,                                      \
                           AsyncClient,                                 \
                           Server,                                      \
                           AsyncServer
from galaxy.net.auth import CredentialBuilder
from galaxy.net.ssh.ssh import SSHAsyncConnectionFactory,               \
                               SSHClientConnection
from galaxy.net.ssh.transport import SSHTransportFactory
from galaxy.net.ssh.protocol import AsyncSSHProtocol
from galaxy.perfo.decorator import timed,                               \
                                   async_timed
from galaxy.kernel.loop import AsyncioLoop


class SFTPClient(Client, ABC):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(SFTPClient, self).__init__()

    @abstractmethod
    def _connect(self) -> None:
        raise NotImplementedError("Should implement _connect()")

    @abstractmethod
    def chdir(self, path: str) -> None:
        raise NotImplementedError("Should implement chdir()")

    @abstractmethod
    def chmod(self, path: str, mode: int) -> None:
        raise NotImplementedError("Should implement chmod()")

    @abstractmethod
    def chown(self, path: str, uid: int, gid: int) -> None:
        raise NotImplementedError("Should implement chown()")

    @abstractmethod
    def get(self, remote_path: str, local_path: str) -> None:
        raise NotImplementedError("Should implement get()")

    @abstractmethod
    def getcwd(self) -> str:
        raise NotImplementedError("Should implement getcwd()")

    @abstractmethod
    def listdir(self, path: str = ".") -> list[str]:
        raise NotImplementedError("Should implement listdir()")

    @abstractmethod
    def mkdir(self, path: str, mode: int | None = 511) -> None:
        raise NotImplementedError("Should implement mkdir()")

    @abstractmethod
    def put(self, local_path: str, remote_path: str) -> None:
        raise NotImplementedError("Should implement put()")

    @abstractmethod
    def rm(self, path: str) -> None:
        raise NotImplementedError("Should implement remove()")

    @abstractmethod
    def mv(self, old_path: str, new_path: str) -> None:
        raise NotImplementedError("Should implement rename()")

    @abstractmethod
    def rmdir(self, path: str, force: bool | None = False) -> None:
        raise NotImplementedError("Should implement rmdir()")

    @abstractmethod
    def _close(self) -> None:
        raise NotImplementedError("Should implement _close()")

    def __repr__(self) -> str:
        return "<SFTPClient(id='{}')>".format(self.id)


class ParamikoSFTPClient(SFTPClient):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(ParamikoSFTPClient, self).__init__()
        self.transport_fact: SSHTransportFactory | None = None
        self.client: sftp_client.SFTPClient | None = None

    @timed
    def _load(self) -> None:
        super(ParamikoSFTPClient, self)._load()

    def _connect(self) -> None:
        opt = CnOpts()
        opt.hostkeys = None
        cred = CredentialBuilder().from_conf(self.conf["cred"]).build()
        transport = self.transport_fact.create()
        transport.connect(username=cred.username,
                          password=cred.password)
        self.client = sftp_client.SFTPClient.from_transport(transport)
        if self.client is None:
            raise

    def chdir(self, path: str) -> None:
        return self.client.chdir(path)

    def chmod(self, path: str, mode: int) -> None:
        return self.client.chmod(path, mode)

    def chown(self, path: str, uid: int, gid: int) -> None:
        return self.client.chown(uid, gid)

    def get(self, remote_path: str, local_path: str) -> None:
        return self.client.get(remote_path, local_path)

    def getcwd(self) -> str:
        return self.client.getcwd()

    def listdir(self, path: str = ".") -> list[str]:
        output = self.client.listdir_attr(path)
        return output

    def mkdir(self, path: str, mode: int | None = 511) -> None:
        return self.client.mkdir(path, mode)

    def put(self, local_path: str, remote_path: str) -> None:
        return self.client.put(local_path, remote_path)

    def rm(self, path: str) -> None:
        return self.client.remove(path)

    def mv(self, old_path: str, new_path: str) -> None:
        return self.client.rename(old_path, new_path)

    def rmdir(self, path: str, force: bool | None = False) -> None:
        if force:
            self._rmdir(path)
        else:
            self.client.rmdir(path)

    def _rmdir(self, path: str) -> None:
        for item in list(self.client.listdir_iter(path)):
            filepath = "/".join([path, item.filename])
            if stat.S_IFMT(item.st_mode) == stat.S_IFDIR:
                self._rmdir(filepath)
                continue
            self.client.remove(filepath)
        return self.client.rmdir(path)

    def _close(self) -> None:
        return self.client.close()

    def __repr__(self) -> str:
        return "<ParamikoSFTPClient(id='{}')>".format(self.id)


class SFTPAsyncClient(AsyncClient, ABC):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(SFTPAsyncClient, self).__init__()
        self.loop: AsyncioLoop | None = None

        # Factory
        self.conn_fact: SSHAsyncConnectionFactory | None = None

    @async_timed
    async def _load(self) -> None:
        await super(SFTPAsyncClient, self)._load()

    @abstractmethod
    async def _connect(self) -> None:
        raise NotImplementedError("Should implement _connect()")

    @abstractmethod
    async def chdir(self, path: str) -> None:
        raise NotImplementedError("Should implement chdir()")

    @abstractmethod
    async def chmod(self, path: str, mode: int) -> None:
        raise NotImplementedError("Should implement chmod()")

    @abstractmethod
    async def chown(self, path: str, uid: int, gid: int) -> None:
        raise NotImplementedError("Should implement chown()")

    @abstractmethod
    async def get(self, remote_path: str, local_path: str) -> None:
        raise NotImplementedError("Should implement get()")

    @abstractmethod
    async def getcwd(self) -> str:
        raise NotImplementedError("Should implement getcwd()")

    @abstractmethod
    async def listdir(self, path: str = ".") -> list[str]:
        raise NotImplementedError("Should implement listdir()")

    @abstractmethod
    async def mkdir(self, path: str, mode: int | None = 511) -> None:
        raise NotImplementedError("Should implement mkdir()")

    @abstractmethod
    async def put(self, local_path: str, remote_path: str) -> None:
        raise NotImplementedError("Should implement put()")

    @abstractmethod
    async def rm(self, path: str) -> None:
        raise NotImplementedError("Should implement remove()")

    @abstractmethod
    async def mv(self, old_path: str, new_path: str) -> None:
        raise NotImplementedError("Should implement rename()")

    @abstractmethod
    async def rmdir(self, path: str, force: bool | None = False) -> None:
        raise NotImplementedError("Should implement rmdir()")

    @abstractmethod
    async def _close(self) -> None:
        raise NotImplementedError("Should implement _close()")

    def __repr__(self) -> str:
        return "<SFTPAsyncClient(id='{}')>".format(self.id)


class AsyncSSHSFTPAsyncClient(SFTPAsyncClient):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(AsyncSSHSFTPAsyncClient, self).__init__()
        self.protocol: AsyncSSHProtocol | None = None
        self.transport: SSHClientConnection | None = None
        self.client: asyncssh.sftp.SFTPClient | None = None

    @async_timed
    async def _load(self) -> None:
        await super(AsyncSSHSFTPAsyncClient, self)._load()

    @async_timed
    async def _connect(self) -> None:
        self.transport, self.protocol = await self.conn_fact.from_conf(self.conf).create(self)
        self.client = await self.transport.start_sftp_client()
        self.transport.set_keepalive(interval=self.conf["keepalive"])

    @async_timed
    async def _exec(self, attempt_nb: int, func: Callable[[...], Awaitable[Any]], *args, **kwargs) -> Any:
        self.log.logger.debug("Execute the function '{}' (Attempt {})".format(getattr(func, "__name__", repr(callable)),
                                                                              attempt_nb))
        try:
            resp = await func(*args, **kwargs)
            self.log.logger.debug("The function '{}' (Attempt {}) has been executed successfully".format(getattr(func, "__name__", repr(callable)),
                                                                                                         attempt_nb))
        except Exception as e:
            self.log.logger.error("The execution of the function '{}' (Attempt {}) fails with the exception '{}'".format(getattr(func, "__name__", repr(callable)),
                                                                                                                         attempt_nb,
                                                                                                                         str(e)))
            raise e
        return resp

    @async_timed
    async def _exec_with_retries(self, func: Callable[[...], Awaitable[Any]], *args, **kwargs) -> Any:
        resp = None
        tenacity_kwargs = {
                           "retry": (retry_if_exception_type(asyncio.TimeoutError) |
                                     retry_if_exception_type(asyncio.CancelledError)),
                           "before": before_log(self.log.logger, logging.DEBUG),
                           "after": after_log(self.log.logger, logging.DEBUG),
                           "reraise": True
                          }
        if self.is_connected:
            if "retries" in self.conf:
                retries = self.conf["retries"].get("total", 1)
                tenacity_kwargs["stop"] = stop_after_attempt(retries)
        if "req_interval" in self.conf:
            tenacity_kwargs["wait"] = wait_fixed(self.conf["req_interval"])
        try:
            async for attempt in AsyncRetrying(**tenacity_kwargs):
                with attempt:
                    resp = await self._exec(attempt.retry_state.attempt_number, func, *args, **kwargs)
        except RetryError:
            pass
        return resp

    async def exec(self, func: Callable[[...], Awaitable[Any]], *args, **kwargs):
        tenacity_kwargs = {
                           "retry": (retry_if_exception_type(ConnectionLost) |
                                     retry_if_exception_type(SFTPNoConnection)),
                           "before": before_log(self.log.logger, logging.DEBUG),
                           "after": after_log(self.log.logger, logging.DEBUG)
                          }
        if "reconnect_interval" in self.conf:
            tenacity_kwargs["wait"] = wait_fixed(self.conf["reconnect_interval"])
        reconnecting = False
        try:
            async for attempt in AsyncRetrying(**tenacity_kwargs):
                resp = None
                with attempt:
                    if reconnecting:
                        await self.connect()
                        reconnecting = False
                    try:
                        resp = await self._exec_with_retries(func, *args, **kwargs)
                    except (ConnectionLost, SFTPNoConnection) as e:
                        if self.state != constant.STATE_CLOSED:
                            if "reconnect_interval" in self.conf:
                                self.log.logger.error("The execution of function '{}' fails : starting to reconnect after {} sec".format(getattr(func, "__name__", repr(callable)),
                                                                                                                                         self.conf["reconnect_interval"]))
                            else:
                                self.log.logger.error("The execution of function '{}' fails : starting to reconnect".format(getattr(func, "__name__", repr(callable))))
                            await self.close()
                            reconnecting = True
                        raise e
        except RetryError:
            pass
        return resp

    async def chdir(self, path: str) -> None:
        await self.exec(self.client.chdir, path)

    async def chmod(self, path: str, mode: int) -> None:
        await self.exec(self.client.chmod, path, mode)

    async def chown(self, path: str, uid: int, gid: int) -> None:
        await self.exec(self.client.chown, path, uid, gid)

    async def get(self, remote_path: str, local_path: str) -> None:
        await self.exec(self.client.get, remote_path, local_path)

    async def getcwd(self) -> str:
        return await self.exec(self.client.getcwd)

    async def listdir(self, path: str = ".") -> list[str]:
        return await self.exec(self.client.listdir, path)

    async def mkdir(self, path: str, mode: int | None = 511) -> None:
        attr = sftp.SFTPAttrs()
        attr.permissions = mode
        await self.exec(self.client.mkdir, path, attr)

    async def put(self, local_path: str, remote_path: str) -> None:
        await self.exec(self.client.put, local_path, remote_path)

    async def rm(self, path: str) -> None:
        await self.exec(self.client.remove, path)

    async def mv(self, old_path: str, new_path: str) -> None:
        await self.exec(self.client.rename, old_path, new_path)

    async def rmdir(self, path: str, force: bool | None = False) -> None:
        await self.exec(self.client.rmdir, path)

    @async_timed
    async def _close(self) -> None:
        self.client.exit()
        await self.client.wait_closed()
        self.transport.close()
        await self.transport.wait_closed()


class ParamikoServerInterface(ServerInterface):
    """
    classdocs
    """

    def __init__(self):
        """
        Constructor
        """
        super(ParamikoServerInterface, self).__init__()
        self.event = threading.Event()

    def check_channel_request(self, kind, chanid):
        if kind == "session":
            return paramiko.OPEN_SUCCEEDED
        return paramiko.OPEN_FAILED_ADMINISTRATIVELY_PROHIBITED

    def check_auth_password(self, username, password):
        if (username == "robey") and (password == "foo"):
            return paramiko.AUTH_SUCCESSFUL
        return paramiko.AUTH_FAILED

    def check_auth_publickey(self, username, key):
        return paramiko.AUTH_SUCCESSFUL

    def check_auth_gssapi_with_mic(self, username, gss_authenticated=paramiko.AUTH_FAILED, cc_file = None):

        if gss_authenticated == paramiko.AUTH_SUCCESSFUL:
            return paramiko.AUTH_SUCCESSFUL
        return paramiko.AUTH_FAILED

    def check_auth_gssapi_keyex(self, username, gss_authenticated=paramiko.AUTH_FAILED, cc_file = None):
        if gss_authenticated == paramiko.AUTH_SUCCESSFUL:
            return paramiko.AUTH_SUCCESSFUL
        return paramiko.AUTH_FAILED

    def enable_auth_gssapi(self):
        return True

    def get_allowed_auths(self, username):
        return "gssapi-keyex,gssapi-with-mic,password,publickey"

    def check_channel_shell_request(self, channel):
        self.event.set()
        return True

    def check_channel_pty_request(self,
                                  channel,
                                  term,
                                  width,
                                  height,
                                  pixelwidth,
                                  pixelheight,
                                  modes):
        return True


class SFTPServer(Server, ABC):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(SFTPServer, self).__init__()

    @abstractmethod
    def _start(self) -> None:
        raise NotImplementedError("Should implement _start()")

    @abstractmethod
    def _stop(self) -> None:
        raise NotImplementedError("Should implement _stop()")

    def __repr__(self) -> str:
        return "<SFTPServer(id='{}')>".format(self.id)


class ParamikoSFTPServer(SFTPServer):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(ParamikoSFTPServer, self).__init__()
        self.event = threading.Event()
        self.host_key = paramiko.RSAKey(filename="test_rsa.key")

    def _start(self) -> None:
        # now connect
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            sock.bind(("", 2200))
        except Exception as e:
            print("*** Bind failed: " + str(e))
            traceback.print_exc()
            sys.exit(1)

        try:
            sock.listen(100)
            print("Listening for connection ...")
            client, addr = sock.accept()
        except Exception as e:
            print("*** Listen/accept failed: " + str(e))
            traceback.print_exc()
            sys.exit(1)

        print("Got a connection!")

        try:
            t = paramiko.Transport(client, gss_kex=True)
            t.set_gss_host(socket.getfqdn(""))
            try:
                t.load_server_moduli()
            except:
                print("(Failed to load moduli -- gex will be unsupported.)")
                raise
            t.add_server_key(self.host_key)
            server = ParamikoServerInterface()
            try:
                t.start_server(server=server)
            except paramiko.SSHException:
                print("*** SSH negotiation failed.")
                sys.exit(1)

            # wait for auth
            chan = t.accept(20)
            if chan is None:
                print("*** No channel.")
                sys.exit(1)
            print("Authenticated!")

            server.event.wait(10)
            if not server.event.is_set():
                print("*** Client never asked for a shell.")
                sys.exit(1)

            chan.send("\r\n\r\nWelcome to my dorky little BBS!\r\n\r\n")
            chan.send("We are on fire all the time!  Hooray!  Candy corn for everyone!\r\n")
            chan.send("Happy birthday to Robot Dave!\r\n\r\n")
            chan.send("Username: ")
            f = chan.makefile("rU")
            username = f.readline().strip("\r\n")
            chan.send("\r\nI don't like you, " + username + ".\r\n")
            chan.close()

        except Exception as e:
            print("*** Caught exception: " + str(e.__class__) + ": " + str(e))
            traceback.print_exc()
            try:
                t.close()
            except:
                pass
            sys.exit(1)

    def _stop(self) -> None:
        pass

    def __repr__(self) -> str:
        return "<ParamikoSFTPServer(id='{}')>".format(self.id)


class SFTPAsyncServer(AsyncServer, ABC):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(SFTPAsyncServer, self).__init__()

    @abstractmethod
    async def _start(self) -> None:
        raise NotImplementedError("Should implement _start()")

    @abstractmethod
    async def _stop(self) -> None:
        raise NotImplementedError("Should implement _stop()")

    def __repr__(self) -> str:
        return "<SFTPAsyncServer(id='{}')>".format(self.id)


class AsyncSSHSFTPServer(sftp.SFTPServer):
    """
    classdocs
    """

    def __init__(self, channel: SSHServerChannel) -> None:
        """
        Constructor
        """
        root = "/tmp/sftp/" + channel.get_extra_info("username")
        os.makedirs(root, exist_ok=True)
        super(AsyncSSHSFTPServer, self).__init__(channel, chroot=root)


class AsyncSSHSFTPAsyncServer(SFTPAsyncServer):
    """
    classdocs
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super(AsyncSSHSFTPAsyncServer, self).__init__()

    async def _start(self) -> None:
        await listen("",
                     8022,
                     server_host_keys=["ssh_host_key"],
                     authorized_client_keys="ssh_user_ca",
                     sftp_factory=AsyncSSHSFTPServer)

    async def _stop(self) -> None:
        pass

    def __repr__(self) -> str:
        return "<SFTPAsyncServer(id='{}')>".format(self.id)
